# Copyright (c) 2015 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import os
import testtools

from murano_tempest_tests.tests.api.service_broker import base
from murano_tempest_tests import utils


class ServiceBrokerActionsTest(base.BaseServiceBrokerAdminTest):

    @testtools.testcase.attr('gate')
    def test_applications_listing(self):
        app_list = self.service_broker_client.get_applications_list()
        self.assertIsInstance(app_list, list)

    @testtools.testcase.attr('smoke')
    @testtools.testcase.attr('gate')
    def test_provision_and_deprovision(self):
        application_name = utils.generate_name('cfapi')
        abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        self.addCleanup(os.remove, abs_archive_path)
        package = self.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': True})
        self.addCleanup(self.application_catalog_client.delete_package,
                        package['id'])
        app_list = self.service_broker_client.get_applications_list()
        app = self.service_broker_client.get_application(application_name,
                                                         app_list)
        post_json = {}
        instance_id = utils.generate_uuid()
        space_id = utils.generate_uuid()
        service = self.service_broker_client.provision(
            instance_id, app['id'], app['plans'][0]['id'],
            space_id, post_json)
        self.wait_for_result(instance_id, 30)
        self.addCleanup(self.perform_deprovision, instance_id)
        self.assertIsInstance(json.loads(service), dict)

    @testtools.testcase.attr('smoke')
    @testtools.testcase.attr('gate')
    def test_binding_instance(self):
        application_name = utils.generate_name('cfapi')
        abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        self.addCleanup(os.remove, abs_archive_path)
        package = self.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': True})
        self.addCleanup(self.application_catalog_client.delete_package,
                        package['id'])
        app_list = self.service_broker_client.get_applications_list()
        app = self.service_broker_client.get_application(application_name,
                                                         app_list)
        post_json = {}
        instance_id = utils.generate_uuid()
        space_id = utils.generate_uuid()
        service = self.service_broker_client.provision(
            instance_id, app['id'], app['plans'][0]['id'],
            space_id, post_json)
        self.wait_for_result(instance_id, 30)
        self.addCleanup(self.perform_deprovision, instance_id)
        self.assertIsInstance(json.loads(service), dict)
        binding = self.service_broker_client.create_binding(instance_id)
        self.assertIsInstance(binding, dict)
        self.assertEqual({'uri': 'localhost'}, binding)

    @testtools.testcase.attr('smoke')
    @testtools.testcase.attr('gate')
    def test_provision_with_incorrect_input(self):
        """Test provision with restricted items in object model

        Test will fail on deprovision, if parameters from '?' section
        will passed through service-broker.
        """
        application_name = utils.generate_name('cfapi')
        abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        self.addCleanup(os.remove, abs_archive_path)
        package = self.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': True})
        self.addCleanup(self.application_catalog_client.delete_package,
                        package['id'])
        app_list = self.service_broker_client.get_applications_list()
        app = self.service_broker_client.get_application(application_name,
                                                         app_list)

        # NOTE(freerunner): The '?' section should be cutted off during
        # provision action.
        post_json = {
            '?': {
                'type': 'io.murano.apps.{0}'.format(application_name),
                'id': utils.generate_uuid()
            }
        }
        instance_id = utils.generate_uuid()
        space_id = utils.generate_uuid()
        service = self.service_broker_client.provision(
            instance_id, app['id'], app['plans'][0]['id'],
            space_id, post_json)
        self.wait_for_result(instance_id, 30)
        self.addCleanup(self.perform_deprovision, instance_id)
        self.assertIsInstance(json.loads(service), dict)

    @testtools.testcase.attr('smoke')
    @testtools.testcase.attr('gate')
    def test_double_provision_to_the_same_space(self):
        application_name = utils.generate_name('cfapi')
        abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        self.addCleanup(os.remove, abs_archive_path)
        package = self.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': True})
        self.addCleanup(self.application_catalog_client.delete_package,
                        package['id'])
        app_list = self.service_broker_client.get_applications_list()
        app = self.service_broker_client.get_application(application_name,
                                                         app_list)
        post_json = {}
        instance_id = utils.generate_uuid()
        space_id = utils.generate_uuid()
        service = self.service_broker_client.provision(
            instance_id, app['id'], app['plans'][0]['id'],
            space_id, post_json)
        self.wait_for_result(instance_id, 30)
        self.addCleanup(self.perform_deprovision, instance_id)
        self.assertIsInstance(json.loads(service), dict)
        application_name = utils.generate_name('cfapi')
        abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        self.addCleanup(os.remove, abs_archive_path)
        package = self.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': True})
        self.addCleanup(self.application_catalog_client.delete_package,
                        package['id'])
        app_list = self.service_broker_client.get_applications_list()
        app = self.service_broker_client.get_application(application_name,
                                                         app_list)
        post_json = {}
        instance_id = utils.generate_uuid()
        service = self.service_broker_client.provision(
            instance_id, app['id'], app['plans'][0]['id'],
            space_id, post_json)
        self.wait_for_result(instance_id, 30)
        self.addCleanup(self.perform_deprovision, instance_id)
        self.assertIsInstance(json.loads(service), dict)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import os
import shutil
import tempfile

import mock
from oslo_config import cfg
import semantic_version
import testtools

from murano.dsl import murano_package as dsl_package
from murano.engine import package_loader
from murano.tests.unit import base
from murano_tempest_tests import utils

CONF = cfg.CONF


class TestPackageCache(base.MuranoTestCase):

    def setUp(self):
        super(TestPackageCache, self).setUp()

        self.location = tempfile.mkdtemp()
        CONF.set_override('enable_packages_cache', True, 'engine')
        self.old_location = CONF.engine.packages_cache
        CONF.set_override('packages_cache', self.location, 'engine')

        self.murano_client = mock.MagicMock()
        package_loader.ApiPackageLoader.client = self.murano_client
        self.loader = package_loader.ApiPackageLoader(None)

    def tearDown(self):
        CONF.set_override('packages_cache', self.old_location, 'engine')
        shutil.rmtree(self.location, ignore_errors=True)
        super(TestPackageCache, self).tearDown()

    @testtools.skipIf(os.name == 'nt', "Doesn't work on Windows")
    def test_load_package(self):
        fqn = 'io.murano.apps.test'
        path, name = utils.compose_package(
            'test',
            os.path.join(self.location, 'manifest.yaml'),
            self.location, archive_dir=self.location)
        with open(path, 'rb') as f:
            package_data = f.read()
        spec = semantic_version.Spec('*')

        first_id, second_id, third_id = '123', '456', '789'
        package = mock.MagicMock()
        package.fully_qualified_name = fqn
        package.id = first_id
        package.version = '0.0.1'

        self.murano_client.packages.filter = mock.MagicMock(
            return_value=[package])
        self.murano_client.packages.download = mock.MagicMock(
            return_value=package_data)

        # load the package
        self.loader.load_class_package(fqn, spec)

        # assert that everything got created
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, first_id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, first_id, 'manifest.yaml')))

        # assert, that we called download
        self.assertEqual(self.murano_client.packages.download.call_count, 1)

        # now that the cache is in place, call it for the 2d time
        self.loader._package_cache = {}
        self.loader._class_cache = {}
        self.loader.load_class_package(fqn, spec)

        # check that we didn't download a thing
        self.assertEqual(self.murano_client.packages.download.call_count, 1)

        # changing id, new package would be downloaded.
        package.id = second_id
        self.loader._package_cache = {}
        self.loader._class_cache = {}
        self.loader.load_class_package(fqn, spec)

        # check that we didn't download a thing
        self.assertEqual(self.murano_client.packages.download.call_count, 2)

        # check that old directories were not deleted
        # we did not call cleanup and did not release the locks
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, first_id)))

        # check that new directories got created correctly
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, second_id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, second_id, 'manifest.yaml')))

        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version)))
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, second_id)))

        # changing id, new package would be downloaded.
        package.id = third_id
        self.loader._package_cache = {}
        self.loader._class_cache = {}

        # release all the locks
        self.loader.cleanup()
        self.loader.load_class_package(fqn, spec)

        # check that we didn't download a thing
        self.assertEqual(self.murano_client.packages.download.call_count, 3)

        # check that old directories were *deleted*
        self.assertFalse(os.path.isdir(os.path.join(
            self.location, fqn, package.version, first_id)))
        self.assertFalse(os.path.isdir(os.path.join(
            self.location, fqn, package.version, second_id)))

        # check that new directories got created correctly
        self.assertTrue(os.path.isdir(os.path.join(
            self.location, fqn, package.version, third_id)))
        self.assertTrue(os.path.isfile(os.path.join(
            self.location, fqn, package.version, third_id, 'manifest.yaml')))


class TestCombinedPackageLoader(base.MuranoTestCase):
    @classmethod
    def setUpClass(cls):
        super(TestCombinedPackageLoader, cls).setUpClass()

        location = os.path.dirname(__file__)
        CONF.set_override('load_packages_from', [location], 'engine',
                          enforce_type=True)
        cls.execution_session = mock.MagicMock()
        cls.loader = package_loader.CombinedPackageLoader(
            cls.execution_session)
        cls.api_loader = mock.MagicMock()
        cls.loader.api_loader = cls.api_loader

        cls.local_pkg_name = 'io.murano.test.MyTest'
        cls.api_pkg_name = 'test.mpl.v1.app.Thing'

    def test_loaders_initialized(self):
        self.assertEqual(1, len(self.loader.directory_loaders),
                         'One directory class loader should be initialized'
                         ' since there is one valid murano pl package in the'
                         ' provided directory in config.')
        self.assertIsInstance(self.loader.directory_loaders[0],
                              package_loader.DirectoryPackageLoader)

    def test_get_package_by_class_directory_loader(self):
        spec = semantic_version.Spec('*')
        result = self.loader.load_class_package(self.local_pkg_name, spec)
        self.assertIsInstance(result, dsl_package.MuranoPackage)

    def test_get_package_by_name_directory_loader(self):
        spec = semantic_version.Spec('*')
        result = self.loader.load_package(self.local_pkg_name, spec)
        self.assertIsInstance(result, dsl_package.MuranoPackage)

    def test_get_package_by_class_api_loader(self):
        spec = semantic_version.Spec('*')
        self.loader.load_package(self.api_pkg_name, spec)

        self.api_loader.load_package.assert_called_with(
            self.api_pkg_name, spec)

    def test_get_package_api_loader(self):
        spec = semantic_version.Spec('*')
        self.loader.load_class_package(self.api_pkg_name, spec)

        self.api_loader.load_class_package.assert_called_with(
            self.api_pkg_name, spec)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import six
import testtools

from tempest import config

from murano_tempest_tests.tests.api.application_catalog.artifacts import base
from murano_tempest_tests import utils

CONF = config.CONF


class TestVersioning(base.BaseArtifactsTest):

    @classmethod
    def resource_setup(cls):
        if not CONF.application_catalog.glare_backend:
            msg = ("Murano is not using GLARE backend. "
                   "Skipping GLARE tests.")
            raise cls.skipException(msg)
        super(TestVersioning, cls).resource_setup()

        application_name = utils.generate_name('package_test')
        # create first package
        version1 = '1.0.0'
        package1, _ = cls.upload_package(application_name,
                                         version=version1)

        # create second package
        version2 = '2.0.0'
        package2, path1 = cls.upload_package(application_name,
                                             version=version2)

        # create package with require >=2.0.0 for 2.0.0 package
        expected_version = '>=2.0.0'
        main_app_name = utils.generate_name('main_package_test')
        require = [(package2['name'], expected_version)]
        package3, path2 = cls.upload_package(main_app_name, require=require)

        cls.packages = {
            '1.0.0': package1,
            '2.0.0': package2,
            'require_for_2.0.0': package3,
        }
        cls.abs_archive_paths = [path1, path2]

    @classmethod
    def resource_cleanup(cls):
        for pkg in six.itervalues(cls.packages):
            cls.artifacts_client.delete_package(pkg['id'])
        map(os.remove, cls.abs_archive_paths)
        super(TestVersioning, cls).resource_cleanup()

    @testtools.testcase.attr('smoke')
    def test_availability_of_packages_with_different_versions(self):
        """Test availability of packages with different versions.

        1) Check two packages to have the same names.
        2) Check two packages to have different ids.
        3) Check two packages to be in repository.
        """
        self.assertEqual(self.packages['1.0.0']['name'],
                         self.packages['2.0.0']['name'])
        self.assertNotEqual(self.packages['1.0.0']['id'],
                            self.packages['2.0.0']['id'])

        # check packages availability
        artifact_packages = {pkg['id'] for pkg in
                             self.artifacts_client.get_list_packages()}

        self.assertIn(self.packages['1.0.0']['id'], artifact_packages)
        self.assertIn(self.packages['2.0.0']['id'], artifact_packages)

    @testtools.testcase.attr('smoke')
    def test_deploy_packages_with_different_versions(self):
        """Test deployment of packages with different versions.

        1) Create environment.
        2) Add package with version 1.0.0 to the environment.
        3) Add package with version 2.0.0 to the environment.
        4) Deploy environment.
        5) Check if deployment status ok.
        """

        # create environment
        environment_name = utils.generate_name('create_environment')
        environment = self.application_catalog_client.create_environment(
            environment_name)
        self.addCleanup(self.application_catalog_client.delete_environment,
                        environment['id'])

        # create session
        session = self.application_catalog_client.create_session(
            environment['id'])

        # add first application
        object_model = self.create_obj_model(self.packages['1.0.0'])

        self.application_catalog_client.create_service(
            environment['id'], session['id'], object_model)

        # add second application
        object_model = self.create_obj_model(self.packages['2.0.0'])

        self.application_catalog_client.create_service(
            environment['id'], session['id'], object_model)

        self.application_catalog_client.deploy_session(
            environment['id'], session['id'])

        deploy_result = utils.wait_for_environment_deploy(
            self.application_catalog_client, environment['id'])['status']

        self.assertEqual(deploy_result, 'ready')

    @testtools.testcase.attr('smoke')
    def test_deploy_package_with_required_package_version(self):
        """Test deployment of package which requires package with present version.

        1) Create environment.
        2) Add to the environment package which requires version 2.0.0 of the
        package, which is present with versions 1.0.0 and 2.0.0 in repository.
        3) Deploy environment.
        4) Check if deployment status ok.
        """

        # create environment
        environment_name = utils.generate_name('create_environment')
        environment = self.application_catalog_client.create_environment(
            environment_name)
        self.addCleanup(self.application_catalog_client.delete_environment,
                        environment['id'])

        # create session
        session = self.application_catalog_client.create_session(
            environment['id'])

        object_model = self.create_obj_model(
            self.packages['require_for_2.0.0'])

        self.application_catalog_client.create_service(
            environment['id'], session['id'], object_model)

        self.application_catalog_client.deploy_session(
            environment['id'], session['id'])

        deploy_result = utils.wait_for_environment_deploy(
            self.application_catalog_client, environment['id'])['status']

        self.assertEqual(deploy_result, 'ready')



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

import murano.tests.functional.common.utils as common_utils
import murano.tests.functional.engine.manager as core
from nose.plugins.attrib import attr as tag


class LanguageSupportTest(core.MuranoTestsCore):

    @classmethod
    def setUpClass(cls):
        super(LanguageSupportTest, cls).setUpClass()
        cls.linux = core.CONF.murano.linux_image
        cls.flavor = core.CONF.murano.standard_flavor
        cls.keyname = core.CONF.murano.keyname
        cls.instance_type = core.CONF.murano.instance_type

        try:
            # Upload the Murano test package.
            cls.upload_app('io.murano.conflang.chef.ExampleChef',
                           'ExampleChef', {"tags": ["tag"]})
            cls.upload_app('io.murano.conflang.puppet.ExamplePuppet',
                           'ExamplePuppet', {"tags": ["tag"]})
        except Exception:
            cls.tearDownClass()
            raise

    @classmethod
    def tearDownClass(cls):
        with common_utils.ignored(Exception):
            try:
                cls.purge_uploaded_packages()
            except Exception:
                raise

    def _test_deploy(self, environment_name, package_name, port):
        post_body = {
            "instance": {
                "flavor": self.flavor,
                "image": self.linux,
                "keyname": self.keyname,
                "assignFloatingIp": True,
                'name': environment_name,
                "?": {
                    "type": "io.murano.resources.ConfLangInstance",
                    "id": str(uuid.uuid4())
                },
            },
            "name": environment_name,
            "port": port,
            "?": {
                "type": package_name,
                "id": str(uuid.uuid4())
            }
        }

        environment_name = environment_name + uuid.uuid4().hex[:5]
        environment = self.create_environment(name=environment_name)
        session = self.create_session(environment)
        self.add_service(environment, post_body, session)
        self.deploy_environment(environment, session)
        self.wait_for_environment_deploy(environment)
        self.deployment_success_check(environment, port)

    @tag('gate', 'all', 'coverage')
    def test_deploy_example_chef_example(self):
        self._test_deploy('chefExample',
                          'io.murano.conflang.chef.ExampleChef', 22)

    @tag('gate', 'all', 'coverage')
    def test_deploy_example_puppet_example(self):
        self._test_deploy('puppetExample',
                          "io.murano.conflang.puppet.ExamplePuppet", 22)



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from testtools import matchers

from murano.dsl import schema_generator
from murano.tests.unit.dsl.foundation import runner
from murano.tests.unit.dsl.foundation import test_case


class TestSchemaGeneration(test_case.DslTestCase):
    def setUp(self):
        super(TestSchemaGeneration, self).setUp()
        schema = schema_generator.generate_schema(
            self.package_loader, runner.TestContextManager({}),
            'TestSchema')

        self._class_schema = schema.pop('')
        self._model_builders_schema = schema

    def test_general_structure(self):
        self.assertIn('$schema', self._class_schema)
        self.assertIn('type', self._class_schema)
        self.assertIn('properties', self._class_schema)
        self.assertEqual(
            'http://json-schema.org/draft-04/schema#',
            self._class_schema['$schema'])
        self.assertEqual('object', self._class_schema['type'])

    def _test_simple_property(self, name_or_schema, types):
        if not isinstance(name_or_schema, dict):
            props = self._class_schema['properties']
            self.assertIn(name_or_schema, props)
            schema = props[name_or_schema]
        else:
            schema = name_or_schema
        self.assertIn('type', schema)
        if isinstance(types, list):
            self.assertItemsEqual(schema['type'], types)
        else:
            self.assertEqual(schema['type'], types)
        return schema

    def test_string_property(self):
        self._test_simple_property('stringProperty', ['null', 'string'])

    def test_not_null_string_property(self):
        self._test_simple_property('stringNotNullProperty', 'string')

    def test_int_property(self):
        self._test_simple_property('intProperty', ['null', 'integer'])

    def test_not_null_int_property(self):
        self._test_simple_property('intNotNullProperty', 'integer')

    def test_bool_property(self):
        self._test_simple_property('boolProperty', ['null', 'boolean'])

    def test_not_null_bool_property(self):
        self._test_simple_property('boolNotNullProperty', 'boolean')

    def test_class_property(self):
        schema = self._test_simple_property(
            'classProperty', ['null', 'muranoObject'])
        self.assertEqual('SampleClass1', schema.get('muranoType'))

    def test_template_property(self):
        schema = self._test_simple_property(
            'templateProperty', ['null', 'muranoObject'])
        self.assertEqual('SampleClass1', schema.get('muranoType'))
        self.assertTrue(schema.get('owned'))
        self.assertItemsEqual(
            ['stringProperty'],
            schema.get('excludedProperties'))

    def test_default_property(self):
        schema = self._test_simple_property(
            'defaultProperty', ['null', 'integer'])
        self.assertEqual(999, schema.get('default'))

    def test_list_property(self):
        schema = self._test_simple_property('listProperty', 'array')
        self.assertIn('items', schema)
        items = schema['items']
        self._test_simple_property(items, 'string')

    def test_dict_property(self):
        schema = self._test_simple_property('dictProperty', 'object')
        self.assertIn('properties', schema)
        props = schema['properties']
        self.assertIn('key1', props)
        self._test_simple_property(props['key1'], 'string')
        self.assertIn('key2', props)
        self._test_simple_property(props['key2'], 'string')
        self.assertIn('additionalProperties', schema)
        extra_props = schema['additionalProperties']
        self._test_simple_property(extra_props, ['null', 'integer'])

    def test_complex_property(self):
        schema = self._test_simple_property('complexProperty', 'object')
        self.assertIn('properties', schema)
        self.assertEqual({}, schema['properties'])
        self.assertIn('additionalProperties', schema)
        extra_props = schema['additionalProperties']
        self._test_simple_property(extra_props, 'array')
        self.assertIn('items', extra_props)
        items = extra_props['items']
        self._test_simple_property(items, 'integer')

    def test_minimum_contract(self):
        schema = self._test_simple_property('minimumContract', 'integer')
        self.assertFalse(schema.get('exclusiveMinimum', True))
        self.assertEqual(5, schema.get('minimum'))

    def test_maximum_contract(self):
        schema = self._test_simple_property('maximumContract', 'integer')
        self.assertTrue(schema.get('exclusiveMaximum', False))
        self.assertEqual(15, schema.get('maximum'))

    def test_range_contract(self):
        schema = self._test_simple_property('rangeContract', 'integer')
        self.assertFalse(schema.get('exclusiveMaximum', True))
        self.assertTrue(schema.get('exclusiveMinimum', False))
        self.assertEqual(0, schema.get('minimum'))
        self.assertEqual(10, schema.get('maximum'))

    def test_chain_contract(self):
        schema = self._test_simple_property('chainContract', 'integer')
        self.assertFalse(schema.get('exclusiveMaximum', True))
        self.assertTrue(schema.get('exclusiveMinimum', False))
        self.assertEqual(0, schema.get('minimum'))
        self.assertEqual(10, schema.get('maximum'))

    def test_regex_contract(self):
        schema = self._test_simple_property('regexContract', 'string')
        self.assertEqual(r'\d+', schema.get('pattern'))

    def test_enum_contract(self):
        schema = self._test_simple_property('enumContract', 'string')
        self.assertEqual(['a', 'b'], schema.get('enum'))

    def test_enum_func_contract(self):
        schema = self._test_simple_property('enumFuncContract', 'string')
        self.assertEqual(['x', 'y'], schema.get('enum'))

    def test_ui_hints(self):
        schema = self._test_simple_property('decoratedProperty', 'string')
        self.assertEqual('Title!', schema.get('title'))
        self.assertEqual('Description!', schema.get('description'))
        self.assertEqual('Help!', schema.get('helpText'))
        self.assertFalse(schema.get('visible'))
        self.assertThat(schema.get('formIndex'), matchers.GreaterThan(-1))
        self.assertEqual('mySection', schema.get('formSection'))

        sections = self._class_schema.get('formSections')
        self.assertIsInstance(sections, dict)
        section = sections.get('mySection')
        self.assertIsInstance(section, dict)
        self.assertThat(section.get('index'), matchers.GreaterThan(-1))
        self.assertEqual('Section Title', section.get('title'))

    def test_model_builders(self):
        self.assertEqual(1, len(self._model_builders_schema))
        schema = self._model_builders_schema.get('modelBuilder')
        self.assertIsInstance(schema, dict)
        self._class_schema = schema
        self.test_general_structure()
        self.assertEqual('Model Builder!', schema.get('title'))
        args = schema['properties']
        self._test_simple_property(args.get('arg1'), 'string')
        self._test_simple_property(args.get('arg2'), 'integer')
        arg1 = args['arg1']
        self.assertEqual('Arg1!', arg1.get('title'))



# Copyright (c) 2015 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from nose.plugins.attrib import attr as tag

import murano.tests.functional.engine.manager as core


class MuranoDeploymentTest(core.MuranoTestsCore):

    @classmethod
    def setUpClass(cls):
        super(MuranoDeploymentTest, cls).setUpClass()

        cls.linux = core.CONF.murano.linux_image
        cls.flavor = core.CONF.murano.standard_flavor

        cls.upload_app('io.murano.apps.test.UpdateExecutor',
                       'UpdateExecutor',
                       {"categories": ["Web"], "tags": ["tag"]})

        cls.upload_app('io.murano.apps.test.Lighttpd',
                       'Lighttpd',
                       {"categories": ["Web"], "tags": ["tag"]})

        cls.upload_app('io.murano.apps.test.ApacheHttpServerCustom',
                       'Apache HTTP Server Custom',
                       {"categories": ["Web"], "tags": ["test"]})

    @classmethod
    def tearDownClass(cls):
        super(MuranoDeploymentTest, cls).tearDownClass()

        cls.purge_environments()
        cls.purge_uploaded_packages()

    @tag('gate', 'all', 'coverage')
    def test_app_deployment(self):
        post_body = self.get_test_app()
        environment_name = self.rand_name('dummyMurano')
        environment = self.create_environment(name=environment_name)
        session = self.create_session(environment)
        self.add_service(environment, post_body, session)
        self.deploy_environment(environment, session)

    @tag('gate', 'all', 'coverage')
    def test_resources_deallocation(self):
        app_1 = self.get_test_app()
        app_2 = self.get_test_app()
        environment_name = self.rand_name('dummyMurano')
        environment = self.create_environment(name=environment_name)
        session = self.create_session(environment)
        self.add_service(environment, app_1, session)
        self.add_service(environment, app_2, session)
        self.deploy_environment(environment, session)

        environment = self.get_environment(environment)
        app_for_remove = self.get_service(environment, app_1['name'],
                                          to_dict=False)
        session = self.create_session(environment)
        environment = self.delete_service(environment, session, app_for_remove)
        self.deploy_environment(environment, session)

        instance_name = app_1['instance']['name']
        stack = self._get_stack(environment.id)
        template = self.get_stack_template(stack)
        ip_addresses = '{0}-assigned-ip'.format(instance_name)
        floating_ip = '{0}-FloatingIPaddress'.format(instance_name)

        self.assertNotIn(ip_addresses, template['outputs'])
        self.assertNotIn(floating_ip, template['outputs'])
        self.assertNotIn(instance_name, template['resources'])

    @tag('gate', 'all', 'coverage')
    def test_dependent_apps(self):
        post_body = self.get_test_app()
        environment_name = self.rand_name('dummyMurano')
        environment = self.create_environment(name=environment_name)
        session = self.create_session(environment)
        updater = self.add_service(environment, post_body, session,
                                   to_dict=True)
        post_body = {
            "name": self.rand_name("lighttest"),
            "updater": updater,
            "?": {
                "type": "io.murano.apps.test.Lighttpd",
                "id": str(uuid.uuid4())
            }
        }
        self.add_service(environment, post_body, session)
        self.deploy_environment(environment, session)
        self.status_check(environment,
                          [[updater['instance']['name'], 22, 80]])

    @tag('gate', 'all', 'coverage')
    def test_simple_software_configuration(self):
        post_body = {
            "instance": {
                "flavor": self.flavor,
                "image": self.linux,
                "assignFloatingIp": True,
                "?": {
                    "type": "io.murano.resources.LinuxMuranoInstance",
                    "id": str(uuid.uuid4())
                },
                "name": self.rand_name("mrn-test"),
            },
            "name": self.rand_name("ssc-test"),
            "userName": self.rand_name("user"),
            "?": {
                "type": "io.murano.apps.test.ApacheHttpServerCustom",
                "id": str(uuid.uuid4())
            }
        }
        username = post_body["userName"]
        environment_name = self.rand_name('SSC-murano')
        environment = self.create_environment(name=environment_name)
        session = self.create_session(environment)
        self.add_service(environment, post_body, session, to_dict=True)
        self.deploy_environment(environment, session)
        self.status_check(environment,
                          [[post_body['instance']['name'], 22, 80]])
        resp = self.check_path(environment, '', post_body['instance']['name'])
        self.assertIn(username, resp.text, "Required information not found in "
                                           "response from server")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Revision ID: 010
Revises: None
Create Date: 2015-09-08 00:00:00.698760

"""

# revision identifiers, used by Alembic.
revision = '010'
down_revision = '009'

from alembic import op
import sqlalchemy as sa


MYSQL_ENGINE = 'InnoDB'
MYSQL_CHARSET = 'utf8'


def upgrade():
    with op.batch_alter_table("environment") as batch_op:
        batch_op.drop_column('networking')
    with op.batch_alter_table("environment-template") as batch_op2:
        batch_op2.drop_column('networking')


def downgrade():
    op.add_column('environment', sa.Column('networking', sa.Text(),
                  nullable=True))
    op.add_column('environment-template', sa.Column('networking', sa.Text(),
                  nullable=True))



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import os.path

import unittest2 as unittest
import yaml

from murano.common import uuidutils
from murano.dsl import helpers
from murano.dsl import package_loader
import murano.policy.congress_rules as congress

TENANT_ID = 'de305d5475b4431badb2eb6b9e546013'


class MockPackageLoader(package_loader.MuranoPackageLoader):
    def __init__(self, rules):
        """Create rules like this: ['child->parent', 'child->parent2']."""

        self._classes = {}
        rules_dict = {}
        for rule in rules:
            split = rule.split('->')
            rules_dict.setdefault(split[0], []).append(split[1])
        classes = (self.get_class(cls, rules_dict) for cls in rules_dict)
        self._package = MockPackage(classes)

    def get_class(self, name, rules_dict):
        if name in self._classes:
            return self._classes[name]
        parents = [self.get_class(parent, rules_dict)
                   for parent in rules_dict.get(name, [])]
        result = MockClass({'name': name, 'declared_parents': parents})
        self._classes[name] = result
        return result

    def register_package(self, package):
        pass

    def load_class_package(self, class_name, version_spec):
        return self._package

    def load_package(self, package_name, version_spec):
        return self._package

    def export_fixation_table(self):
        pass

    def import_fixation_table(self, fixations):
        pass

    def compact_fixation_table(self):
        pass


class MockPackage(object):
    def __init__(self, classes):
        self._classes = {}
        for cls in classes:
            self._classes[cls.name] = cls

    @property
    def classes(self):
        return self._classes.keys()

    def find_class(self, name, *args, **kwargs):
        return self._classes.get(name)


class MockClass(object):
    def __init__(self, entries):
        self.__dict__.update(entries)

    def ancestors(self):
        return helpers.traverse(self, lambda t: t.declared_parents)


class TestCongressRules(unittest.TestCase):

    def _load_file(self, file_name):
        model_file = os.path.join(
            os.path.dirname(inspect.getfile(self.__class__)), file_name)

        with open(model_file) as stream:
            return yaml.safe_load(stream)

    def _create_rules_str(self, model_file, package_loader=None):
        model = self._load_file(model_file)

        congress_rules = congress.CongressRulesManager()
        rules = congress_rules.convert(model, package_loader,
                                       tenant_id=TENANT_ID)
        rules_str = ", \n".join(map(str, rules))

        return rules_str

    def test_transitive_closure(self):
        closure = congress.CongressRulesManager.transitive_closure(
            [(1, 2), (2, 3), (3, 4)])
        self.assertIn((1, 4), closure)
        self.assertIn((2, 4), closure)

    def test_empty_model(self):
        congress_rules = congress.CongressRulesManager()
        rules = congress_rules.convert(None)
        self.assertEqual(0, len(rules))

    def test_convert_simple_app(self):
        rules_str = self._create_and_check_rules_str('model')

        self.assertNotIn("instance.", rules_str)

    def test_convert_model_two_instances(self):
        rules_str = self._create_and_check_rules_str('model_two_instances')

        self.assertNotIn("\"instances\"", rules_str)

    def test_convert_model_with_relations(self):
        rules_str = self._create_rules_str('model_with_relations.yaml')

        self.assertNotIn(
            'murano:properties+("50fa68ff-cd9a-4845-b573-2c80879d158d", '
            '"server", "8ce94f23-f16a-40a1-9d9d-a877266c315d")', rules_str)

        self.assertIn(
            'murano:relationships+("50fa68ff-cd9a-4845-b573-2c80879d158d", '
            '"8ce94f23-f16a-40a1-9d9d-a877266c315d", "server")', rules_str)

        self.assertIn(
            'murano:relationships+("0aafd67e-72e9-4ae0-bb62-fe724f77df2a", '
            '"ed8df2b0-ddd2-4009-b3c9-2e7a368f3cb8", "instance")', rules_str)

    def test_convert_model_transitive_relationships(self):
        rules_str = self._create_rules_str('model_with_relations.yaml')

        self.assertIn(
            'murano:connected+("50fa68ff-cd9a-4845-b573-2c80879d158d", '
            '"8ce94f23-f16a-40a1-9d9d-a877266c315d")', rules_str)

        self.assertIn(
            'murano:connected+("8ce94f23-f16a-40a1-9d9d-a877266c315d", '
            '"fc6b8c41-166f-4fc9-a640-d82009e0a03d")', rules_str)

    def test_convert_model_services_relationship(self):
        rules_str = self._create_rules_str('model_with_relations.yaml')

        self.assertIn(
            'murano:relationships+("3409bdd0590e4c60b70fda5e6777ff96", '
            '"8ce94f23-f16a-40a1-9d9d-a877266c315d", "services")', rules_str)

        self.assertIn(
            'murano:relationships+("3409bdd0590e4c60b70fda5e6777ff96", '
            '"50fa68ff-cd9a-4845-b573-2c80879d158d", "services")', rules_str)

    def test_convert_model_complex(self):
        self._create_and_check_rules_str('model_complex')

    def test_convert_renamed_app(self):
        self._create_and_check_rules_str('model_renamed')

    def test_parent_types(self):

        #     grand-parent
        #       /     \
        #  parent1   parent2
        #       \     /
        # io.murano.apps.linux.Git

        package_loader = MockPackageLoader([
            'io.murano.apps.linux.Git->parent1',
            'io.murano.apps.linux.Git->parent2',
            'parent1->grand-parent',
            'parent2->grand-parent'
        ])

        rules_str = self._create_rules_str('model.yaml', package_loader)

        self.assertIn(
            'murano:parent_types+("0c810278-7282-4e4a-9d69-7b4c36b6ce6f",'
            ' "parent1")', rules_str)

        self.assertIn(
            'murano:parent_types+("0c810278-7282-4e4a-9d69-7b4c36b6ce6f",'
            ' "parent2")', rules_str)

        self.assertIn(
            'murano:parent_types+("0c810278-7282-4e4a-9d69-7b4c36b6ce6f",'
            ' "grand-parent")', rules_str)

        self.assertIn(
            'murano:parent_types+("0c810278-7282-4e4a-9d69-7b4c36b6ce6f",'
            ' "io.murano.apps.linux.Git")', rules_str)

    def test_to_dictionary(self):
        """test to_dictionary

        If model contains object entry (not dict)
        we try to convert to dict using 'to_dictionary' method.
        """

        class Struct(object):
            def __init__(self, d):
                self.__dict__ = d

            def to_dictionary(self):
                return self.__dict__

            def __getitem__(self, item):
                return self.__dict__[item]

        d = {'?': {'id': '1', 'type': 't1'},
             'apps': [Struct({'?': {'id': '2', 'type': 't2'},
                              'instances': [Struct(
                                  {'?': {'id': '3', 'type': 't3'}})]})]
             }

        model = Struct(d)

        congress_rules = congress.CongressRulesManager()
        tenant_id = uuidutils.generate_uuid()
        rules = congress_rules.convert(model, tenant_id=tenant_id)
        rules_str = ", \n".join(map(str, rules))

        self.assertIn('murano:objects+("1", "{0}", "t1")'.format(tenant_id),
                      rules_str)
        self.assertIn('murano:objects+("2", "1", "t2")', rules_str)
        self.assertIn('murano:objects+("3", "2", "t3")', rules_str)

    def test_environment_owner(self):
        model = self._load_file("model.yaml")
        congress_rules = congress.CongressRulesManager()
        rules = congress_rules.convert(model, tenant_id='tenant1')
        rules_str = ", \n".join(map(str, rules))
        self.assertIn('murano:objects+("c86104748a0c4907b4c5981e6d3bce9f", '
                      '"tenant1", "io.murano.Environment")', rules_str)

    def test_wordpress(self):
        package_loader = MockPackageLoader([
            'io.murano.Environment->io.murano.Object',
            'io.murano.resources.NeutronNetwork->io.murano.resources.Network',
            'io.murano.resources.Network->io.murano.Object',
            'io.murano.databases.MySql->io.murano.databases.SqlDatabase',
            'io.murano.databases.MySql->io.murano.Application',
            'io.murano.databases.SqlDatabase->io.murano.Object',
            'io.murano.resources.LinuxInstance->io.murano.resources.Instance',
            'io.murano.resources.Instance->io.murano.Object',
            'io.murano.Application->io.murano.Object',
            'io.murano.apps.apache.ApacheHttpServer->io.murano.Application',
            'io.murano.apps.ZabbixServer->io.murano.Application',
            'io.murano.apps.ZabbixAgent->io.murano.Application',
            'io.murano.apps.WordPress->io.murano.Application',

            'io.murano.resources.LinuxMuranoInstance->'
            'io.murano.resources.LinuxInstance'
        ])

        self._create_and_check_rules_str('wordpress', package_loader)

    def _create_and_check_rules_str(self, model_name, package_loader=None):
        rules_str = self._create_rules_str(
            '{0}.yaml'.format(model_name), package_loader)
        self._check_expected_rules(rules_str,
                                   'expected_rules_{0}.txt'.format(model_name))
        return rules_str

    def _check_expected_rules(self, rules_str, expected_rules_file_name):
        expected_rules_file = os.path.join(
            os.path.dirname(inspect.getfile(self.__class__)),
            expected_rules_file_name)

        s = ''
        with open(expected_rules_file) as f:
            for line in f:
                line = line.rstrip('\n')
                if line not in rules_str:
                    s += 'Expected rule not found:\n\t' + line + '\n'

        if len(s) > 0:
            self.fail(s)

    def test_state_rule(self):
        rules_str = self._create_rules_str('model.yaml')

        self.assertIn(
            'murano:states+("c86104748a0c4907b4c5981e6d3bce9f", "pending")',
            rules_str)



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import itertools
import os
import os.path
import shutil
import sys
import tempfile
import uuid

import eventlet
from muranoclient.common import exceptions as muranoclient_exc
from muranoclient.glance import client as glare_client
import muranoclient.v1.client as muranoclient
from oslo_config import cfg
from oslo_log import log as logging
from oslo_log import versionutils
import six

from murano.common import auth_utils
from murano.common.i18n import _LE, _LI, _LW
from murano.dsl import constants
from murano.dsl import exceptions
from murano.dsl import helpers
from murano.dsl import package_loader
from murano.engine import murano_package
from murano.engine.system import system_objects
from murano.engine import yaql_yaml_loader
from murano.packages import exceptions as pkg_exc
from murano.packages import load_utils
from murano import utils as m_utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

download_mem_locks = collections.defaultdict(m_utils.ReaderWriterLock)
usage_mem_locks = collections.defaultdict(m_utils.ReaderWriterLock)


class ApiPackageLoader(package_loader.MuranoPackageLoader):
    def __init__(self, execution_session, root_loader=None):
        self._cache_directory = self._get_cache_directory()
        self._class_cache = {}
        self._package_cache = {}
        self._root_loader = root_loader or self
        self._execution_session = execution_session
        self._last_glare_token = None
        self._glare_client = None
        self._murano_client = None
        self._murano_client_session = None

        self._mem_locks = []
        self._ipc_locks = []
        self._downloaded = []
        self._fixations = collections.defaultdict(set)
        self._new_fixations = collections.defaultdict(set)

    def _get_glare_client(self):
        glare_settings = CONF.glare
        session = auth_utils.get_client_session(self._execution_session)
        token = session.auth.get_token(session)
        if self._last_glare_token != token:
            self._last_glare_token = token
            self._glare_client = None

        if self._glare_client is None:
            url = glare_settings.url
            if not url:
                url = session.get_endpoint(
                    service_type='artifact',
                    interface=glare_settings.endpoint_type,
                    region_name=CONF.home_region)

            self._glare_client = glare_client.Client(
                endpoint=url, token=token,
                insecure=glare_settings.insecure,
                key_file=glare_settings.key_file or None,
                ca_file=glare_settings.ca_file or None,
                cert_file=glare_settings.cert_file or None,
                type_name='murano',
                type_version=1)
        return self._glare_client

    @property
    def client(self):
        murano_settings = CONF.murano
        last_glare_client = self._glare_client
        if CONF.engine.packages_service in ['glance', 'glare']:
            if CONF.engine.packages_service == 'glance':
                versionutils.report_deprecated_feature(
                    LOG,
                    _LW("'glance' packages_service option has been renamed "
                        "to 'glare', please update your configuration"))
            artifacts_client = self._get_glare_client()
        else:
            artifacts_client = None
        if artifacts_client != last_glare_client:
            self._murano_client = None
        if not self._murano_client:
            parameters = auth_utils.get_session_client_parameters(
                service_type='application-catalog',
                execution_session=self._execution_session,
                conf=murano_settings
            )
            self._murano_client = muranoclient.Client(
                artifacts_client=artifacts_client, **parameters)
        return self._murano_client

    def load_class_package(self, class_name, version_spec):
        packages = self._class_cache.get(class_name)
        if packages:
            version = version_spec.select(six.iterkeys(packages))
            if version:
                return packages[version]

        filter_opts = {'class_name': class_name,
                       'version': helpers.breakdown_spec_to_query(
                           version_spec)}
        try:
            package_definition = self._get_definition(filter_opts)
            self._lock_usage(package_definition)
        except LookupError:
            exc_info = sys.exc_info()
            six.reraise(exceptions.NoPackageForClassFound,
                        exceptions.NoPackageForClassFound(class_name),
                        exc_info[2])
        return self._to_dsl_package(
            self._get_package_by_definition(package_definition))

    def load_package(self, package_name, version_spec):
        fixed_versions = self._fixations[package_name]
        version = version_spec.select(fixed_versions)
        if version:
            version_spec = helpers.parse_version_spec(version)

        packages = self._package_cache.get(package_name)
        if packages:
            version = version_spec.select(six.iterkeys(packages))
            if version:
                return packages[version]

        filter_opts = {'fqn': package_name,
                       'version': helpers.breakdown_spec_to_query(
                           version_spec)}
        try:
            package_definition = self._get_definition(filter_opts)
            self._lock_usage(package_definition)
        except LookupError:
            exc_info = sys.exc_info()
            six.reraise(exceptions.NoPackageFound(package_name),
                        None, exc_info[2])
        else:
            package = self._get_package_by_definition(package_definition)
            self._fixations[package_name].add(package.version)
            self._new_fixations[package_name].add(package.version)
            return self._to_dsl_package(package)

    def register_package(self, package):
        for name in package.classes:
            self._class_cache.setdefault(name, {})[package.version] = package
        self._package_cache.setdefault(package.name, {})[
            package.version] = package

    @staticmethod
    def _get_cache_directory():
        base_directory = (
            CONF.engine.packages_cache or
            os.path.join(tempfile.gettempdir(), 'murano-packages-cache')
        )

        if CONF.engine.enable_packages_cache:
            directory = os.path.abspath(base_directory)
        else:
            directory = os.path.abspath(os.path.join(base_directory,
                                                     uuid.uuid4().hex))

        if not os.path.isdir(directory):
            m_utils.ensure_tree(directory)

        LOG.debug('Cache for package loader is located at: {dir}'.format(
            dir=directory))
        return directory

    def _get_definition(self, filter_opts):
        filter_opts['catalog'] = True
        try:
            packages = list(self.client.packages.filter(
                **filter_opts))
            if len(packages) > 1:
                LOG.debug('Ambiguous package resolution: more than 1 package '
                          'found for query "{opts}", will resolve based on the'
                          ' ownership'.format(opts=filter_opts))
                return self._get_best_package_match(packages)
            elif len(packages) == 1:
                return packages[0]
            else:
                LOG.debug('There are no packages matching filter '
                          '{opts}'.format(opts=filter_opts))
                raise LookupError()
        except muranoclient_exc.HTTPException:
            LOG.debug('Failed to get package definition from repository')
            raise LookupError()

    def _to_dsl_package(self, app_package):
        dsl_package = murano_package.MuranoPackage(
            self._root_loader, app_package)
        for name in app_package.classes:
            dsl_package.register_class(
                (lambda cls: lambda: get_class(app_package, cls))(name),
                name)
        if app_package.full_name == constants.CORE_LIBRARY:
            system_objects.register(dsl_package)
        self.register_package(dsl_package)
        return dsl_package

    def _get_package_by_definition(self, package_def):
        package_id = package_def.id
        package_directory = os.path.join(
            self._cache_directory,
            package_def.fully_qualified_name,
            getattr(package_def, 'version', '0.0.0'),
            package_id)

        if os.path.isdir(package_directory):
            try:
                return load_utils.load_from_dir(package_directory)
            except pkg_exc.PackageLoadError:
                LOG.exception(
                    _LE('Unable to load package from cache. Clean-up.'))
                shutil.rmtree(package_directory, ignore_errors=True)

        # the package is not yet in cache, let's try and download it.
        download_lock_path = os.path.join(
            self._cache_directory, '{}_download.lock'.format(package_id))
        download_ipc_lock = m_utils.ExclusiveInterProcessLock(
            path=download_lock_path, sleep_func=eventlet.sleep)

        with download_mem_locks[package_id].write_lock(), download_ipc_lock:

            # NOTE(kzaitsev):
            # in case there were 2 concurrent threads/processes one might have
            # already downloaded this package. Check before trying to download
            if os.path.isdir(package_directory):
                try:
                    return load_utils.load_from_dir(package_directory)
                except pkg_exc.PackageLoadError:
                    LOG.error(
                        _LE('Unable to load package from cache. Clean-up.'))
                    shutil.rmtree(package_directory, ignore_errors=True)

            # attempt the download itself
            try:
                LOG.debug("Attempting to download package {} {}".format(
                    package_def.fully_qualified_name, package_id))
                package_data = self.client.packages.download(package_id)
            except muranoclient_exc.HTTPException as e:
                msg = 'Error loading package id {0}: {1}'.format(
                    package_id, str(e)
                )
                exc_info = sys.exc_info()
                six.reraise(pkg_exc.PackageLoadError,
                            pkg_exc.PackageLoadError(msg),
                            exc_info[2])
            package_file = None
            try:
                with tempfile.NamedTemporaryFile(delete=False) as package_file:
                    package_file.write(package_data)

                with load_utils.load_from_file(
                        package_file.name,
                        target_dir=package_directory,
                        drop_dir=False) as app_package:
                    LOG.info(_LI(
                        "Successfully downloaded and unpacked package {} {}")
                        .format(package_def.fully_qualified_name, package_id))
                    self._downloaded.append(app_package)

                    self.try_cleanup_cache(
                        os.path.split(package_directory)[0],
                        current_id=package_id)
                    return app_package
            except IOError:
                msg = 'Unable to extract package data for %s' % package_id
                exc_info = sys.exc_info()
                six.reraise(pkg_exc.PackageLoadError,
                            pkg_exc.PackageLoadError(msg),
                            exc_info[2])
            finally:
                try:
                    if package_file:
                        os.remove(package_file.name)
                except OSError:
                    pass

    def try_cleanup_cache(self, package_directory=None, current_id=None):
        """Attempts to cleanup cache in a given directory.

        :param package_directory: directory containing cached packages
        :param current_id: optional id of the package to exclude from the list
        of deleted packages
        """
        if not package_directory:
            return

        try:
            pkg_ids_listed = set(os.listdir(package_directory))
        except OSError:
            # No directory for this package, probably someone
            # already deleted everything. Anyway nothing to delete
            return

        # if current_id was given: ensure it's not checked for removal
        pkg_ids_listed -= {current_id}

        for pkg_id in pkg_ids_listed:
            stale_directory = os.path.join(
                package_directory,
                pkg_id)

            if not os.path.isdir(package_directory):
                continue

            usage_lock_path = os.path.join(
                self._cache_directory,
                '{}_usage.lock'.format(pkg_id))
            ipc_lock = m_utils.ExclusiveInterProcessLock(
                path=usage_lock_path, sleep_func=eventlet.sleep)

            try:
                with usage_mem_locks[pkg_id].write_lock(False) as acquired:
                    if not acquired:
                        # the package is in use by other deployment in this
                        # process will do nothing, someone else would delete it
                        continue
                    acquired_ipc_lock = ipc_lock.acquire(blocking=False)
                    if not acquired_ipc_lock:
                        # the package is in use by other deployment in another
                        # process, will do nothing, someone else would delete
                        continue

                    shutil.rmtree(stale_directory,
                                  ignore_errors=True)
                    ipc_lock.release()

                    for lock_type in ('usage', 'download'):
                        lock_path = os.path.join(
                            self._cache_directory,
                            '{}_{}.lock'.format(pkg_id, lock_type))
                        try:
                            os.remove(lock_path)
                        except OSError:
                            LOG.warning(
                                _LW("Couldn't delete lock file: "
                                    "{}").format(lock_path))
            except RuntimeError:
                # couldn't upgrade read lock to write-lock. go on.
                continue

    def _get_best_package_match(self, packages):
        public = None
        other = []
        for package in packages:
            if package.owner_id == self._execution_session.project_id:
                return package
            elif package.is_public:
                public = package
            else:
                other.append(package)
        if public is not None:
            return public
        elif other:
            return other[0]

    def _lock_usage(self, package_definition):
        """Locks package for usage"""

        # do not lock anything if we do not persist packages on disk
        if not CONF.engine.enable_packages_cache:
            return

        package_id = package_definition.id

        # A work around the fact that read_lock only supports `with` syntax.
        mem_lock = _with_to_generator(
            usage_mem_locks[package_id].read_lock())

        usage_lock_path = os.path.join(self._cache_directory,
                                       '{}_usage.lock'.format(package_id))
        ipc_lock = m_utils.SharedInterProcessLock(
            path=usage_lock_path,
            sleep_func=eventlet.sleep
        )
        ipc_lock = _with_to_generator(ipc_lock)

        next(mem_lock)
        next(ipc_lock)
        self._mem_locks.append(mem_lock)
        self._ipc_locks.append(ipc_lock)

    def import_fixation_table(self, fixations):
        self._fixations = deserialize_package_fixations(fixations)

    def export_fixation_table(self):
        return serialize_package_fixations(self._fixations)

    def compact_fixation_table(self):
        self._fixations = self._new_fixations.copy()

    def cleanup(self):
        """Cleans up any lock we had acquired and removes any stale packages"""

        if not CONF.engine.enable_packages_cache:
            shutil.rmtree(self._cache_directory, ignore_errors=True)
            return

        for lock in itertools.chain(self._mem_locks, self._ipc_locks):
            try:
                next(lock)
            except StopIteration:
                continue

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()
        return False


class DirectoryPackageLoader(package_loader.MuranoPackageLoader):
    def __init__(self, base_path, root_loader=None):
        self._base_path = base_path
        self._packages_by_class = {}
        self._packages_by_name = {}
        self._loaded_packages = set()
        self._root_loader = root_loader or self
        self._fixations = collections.defaultdict(set)
        self._new_fixations = collections.defaultdict(set)
        self._build_index()

    def _build_index(self):
        for folder in self.search_package_folders(self._base_path):
            try:
                package = load_utils.load_from_dir(folder)
                dsl_package = murano_package.MuranoPackage(
                    self._root_loader, package)
                for class_name in package.classes:
                    dsl_package.register_class(
                        (lambda pkg, cls:
                            lambda: get_class(pkg, cls))(package, class_name),
                        class_name
                    )
                if dsl_package.name == constants.CORE_LIBRARY:
                    system_objects.register(dsl_package)
                self.register_package(dsl_package)
            except pkg_exc.PackageLoadError:
                LOG.info(_LI('Unable to load package from path: {0}').format(
                    folder))
                continue
            LOG.info(_LI('Loaded package from path {0}').format(folder))

    def import_fixation_table(self, fixations):
        self._fixations = deserialize_package_fixations(fixations)

    def export_fixation_table(self):
        return serialize_package_fixations(self._fixations)

    def compact_fixation_table(self):
        self._fixations = self._new_fixations.copy()

    def load_class_package(self, class_name, version_spec):
        packages = self._packages_by_class.get(class_name)
        if not packages:
            raise exceptions.NoPackageForClassFound(class_name)
        version = version_spec.select(six.iterkeys(packages))
        if not version:
            raise exceptions.NoPackageForClassFound(class_name)
        return packages[version]

    def load_package(self, package_name, version_spec):
        fixed_versions = self._fixations[package_name]
        version = version_spec.select(fixed_versions)
        if version:
            version_spec = helpers.parse_version_spec(version)
        packages = self._packages_by_name.get(package_name)
        if not packages:
            raise exceptions.NoPackageFound(package_name)
        version = version_spec.select(six.iterkeys(packages))
        if not version:
            raise exceptions.NoPackageFound(package_name)
        self._fixations[package_name].add(version)
        self._new_fixations[package_name].add(version)
        return packages[version]

    def register_package(self, package):
        for c in package.classes:
            self._packages_by_class.setdefault(c, {})[
                package.version] = package
        self._packages_by_name.setdefault(package.name, {})[
            package.version] = package

    @property
    def packages(self):
        for package_versions in six.itervalues(self._packages_by_name):
            for package in six.itervalues(package_versions):
                yield package

    @staticmethod
    def split_path(path):
        tail = True
        while tail:
            path, tail = os.path.split(path)
            if tail:
                yield path

    @classmethod
    def search_package_folders(cls, path):
        packages = set()
        for folder, _, files in os.walk(path):
            if 'manifest.yaml' in files:
                found = False
                for part in cls.split_path(folder):
                    if part in packages:
                        found = True
                        break
                if not found:
                    packages.add(folder)
                    yield folder

    def cleanup(self):
        """A stub for possible cleanup"""
        pass


class CombinedPackageLoader(package_loader.MuranoPackageLoader):
    def __init__(self, execution_session, root_loader=None):
        root_loader = root_loader or self
        self.api_loader = ApiPackageLoader(execution_session, root_loader)
        self.directory_loaders = []

        for folder in CONF.engine.load_packages_from:
            if os.path.exists(folder):
                self.directory_loaders.append(DirectoryPackageLoader(
                    folder, root_loader))

    def load_package(self, package_name, version_spec):
        for loader in self.directory_loaders:
            try:
                return loader.load_package(package_name, version_spec)
            except exceptions.NoPackageFound:
                continue
        return self.api_loader.load_package(
            package_name, version_spec)

    def load_class_package(self, class_name, version_spec):
        for loader in self.directory_loaders:
            try:
                return loader.load_class_package(class_name, version_spec)
            except exceptions.NoPackageForClassFound:
                continue
        return self.api_loader.load_class_package(
            class_name, version_spec)

    def register_package(self, package):
        self.api_loader.register_package(package)

    def export_fixation_table(self):
        result = deserialize_package_fixations(
            self.api_loader.export_fixation_table())
        for loader in self.directory_loaders:
            fixations = deserialize_package_fixations(
                loader.export_fixation_table())
            for key, value in six.iteritems(fixations):
                result[key].update(value)
        return serialize_package_fixations(result)

    def import_fixation_table(self, fixations):
        self.api_loader.import_fixation_table(fixations)
        for loader in self.directory_loaders:
            loader.import_fixation_table(fixations)

    def compact_fixation_table(self):
        self.api_loader.compact_fixation_table()
        for loader in self.directory_loaders:
            loader.compact_fixation_table()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()
        return False

    def cleanup(self):
        """Calls cleanup method of all loaders we combine"""
        self.api_loader.cleanup()
        for d_loader in self.directory_loaders:
            d_loader.cleanup()


def get_class(package, name):
    version = package.runtime_version
    loader = yaql_yaml_loader.get_loader(version)
    contents, file_id = package.get_class(name)
    return loader(contents, file_id)


def _with_to_generator(context_obj):
    with context_obj as obj:
        yield obj
    yield


def deserialize_package_fixations(fixations):
    result = collections.defaultdict(set)
    for name, versions in six.iteritems(fixations):
        for version in versions:
            result[name].add(helpers.parse_version(version))
    return result


def serialize_package_fixations(fixations):
    return {
        name: list(str(v) for v in versions)
        for name, versions in six.iteritems(fixations)
    }



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from oslo_config import cfg
import yaml

from murano.dsl import murano_package

CONF = cfg.CONF


class MuranoPackage(murano_package.MuranoPackage):
    def __init__(self, package_loader, application_package):
        self.application_package = application_package
        super(MuranoPackage, self).__init__(
            package_loader,
            application_package.full_name,
            application_package.version,
            application_package.runtime_version,
            application_package.requirements,
            application_package.meta
        )

    def get_class_config(self, name):
        version_parts = (
            str(self.version.major),
            str(self.version.minor),
            str(self.version.patch)
        )
        num_parts = len(version_parts)

        for i in range(num_parts + 1):
            for ext in ('json', 'yaml'):
                if i == num_parts:
                    if version_parts[0] == '0':
                        version_suffix = ''
                    else:
                        continue
                else:
                    version_suffix = '-' + '.'.join(
                        version_parts[:num_parts - i])
                file_name = '{name}{version}.{extension}'.format(
                    name=name, version=version_suffix, extension=ext)
                path = os.path.join(CONF.engine.class_configs, file_name)
                if os.path.exists(path):
                    if ext == 'json':
                        with open(path) as f:
                            return json.load(f)
                    else:
                        with open(path) as f:
                            return yaml.safe_load(f)
        return {}

    def get_resource(self, name):
        return self.application_package.get_resource(name)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import six
from yaql.language import expressions
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes

from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.dsl import helpers
from murano.dsl import reflection
from murano.dsl import serializer


@specs.parameter('object_', dsl.MuranoObjectParameter())
def id_(object_):
    return object_.id


@specs.parameter('object_', dsl.MuranoObjectParameter())
@specs.parameter('type__', dsl.MuranoTypeParameter())
@specs.parameter('version_spec', yaqltypes.String(True))
def cast(context, object_, type__, version_spec=None):
    return helpers.cast(
        object_, type__.type.name,
        version_spec or helpers.get_type(context))


@specs.parameter('__type_name', dsl.MuranoTypeParameter())
@specs.parameter('__extra', utils.MappingType)
@specs.parameter('__owner', dsl.MuranoObjectParameter(
    nullable=True, decorate=False))
@specs.parameter('__object_name', yaqltypes.String(True))
def new(__context, __type_name, __owner=None, __object_name=None, __extra=None,
        **parameters):

    data = {
        __type_name: parameters,
        'name': __object_name
    }
    for key, value in six.iteritems(__extra or {}):
        if key.startswith('_'):
            data[key] = value

    object_store = helpers.get_object_store()
    return object_store.load(data, __owner, context=__context,
                             scope_type=helpers.get_names_scope(__context))


@specs.parameter('type_name', dsl.MuranoTypeParameter())
@specs.parameter('parameters', utils.MappingType)
@specs.parameter('extra', utils.MappingType)
@specs.parameter('owner', dsl.MuranoObjectParameter(
    nullable=True, decorate=False))
@specs.parameter('object_name', yaqltypes.String(True))
@specs.name('new')
def new_from_dict(type_name, context, parameters,
                  owner=None, object_name=None, extra=None):
    return new(context, type_name, owner, object_name, extra,
               **utils.filter_parameters_dict(parameters))


@specs.name('new')
@specs.parameter('owner', dsl.MuranoObjectParameter(
    nullable=True, decorate=False))
@specs.parameter('model', utils.MappingType)
def new_from_model(context, model, owner=None):
    object_store = helpers.get_object_store()
    return object_store.load(model, owner, context=context,
                             scope_type=helpers.get_names_scope(context))


@specs.parameter('object_', dsl.MuranoObjectParameter(decorate=False))
@specs.parameter('func', yaqltypes.Lambda())
def super_(context, object_, func=None):
    cast_type = helpers.get_type(context)
    if func is None:
        return [object_.cast(type) for type in cast_type.parents]
    return six.moves.map(func, super_(context, object_))


@specs.parameter('object_', dsl.MuranoObjectParameter(decorate=False))
@specs.parameter('func', yaqltypes.Lambda())
def psuper(context, object_, func=None):
    if func is None:
        return super_(context, object_)
    return helpers.parallel_select(super_(context, object_), func)


@specs.extension_method
def require(value):
    if value is None:
        raise ValueError('Required value is missing')
    return value


@specs.parameter('obj', dsl.MuranoObjectParameter())
@specs.parameter('murano_class_ref', dsl.MuranoTypeParameter())
@specs.extension_method
def find(obj, murano_class_ref):
    return obj.find_owner(murano_class_ref, optional=True)


@specs.parameter('seconds', yaqltypes.Number())
def sleep_(seconds):
    eventlet.sleep(seconds)


@specs.parameter('object_', dsl.MuranoObjectParameter(nullable=True))
def type_(object_):
    return None if object_ is None else object_.type.get_reference()


@specs.name('type')
@specs.parameter('object_', dsl.MuranoObjectParameter(nullable=True))
def type_legacy(object_):
    return None if object_ is None else object_.type.name


@specs.name('type')
@specs.parameter('cls', dsl.MuranoTypeParameter())
def type_from_name(cls):
    return cls


@specs.parameter('object_', dsl.MuranoObjectParameter(nullable=True))
def typeinfo(object_):
    return None if object_ is None else object_.type


@specs.parameter('cls', dsl.MuranoTypeParameter())
@specs.name('typeinfo')
def typeinfo_for_class(cls):
    return cls.type


@specs.parameter('object_', dsl.MuranoObjectParameter(nullable=True))
def name(object_):
    return None if object_ is None else object_.name


@specs.parameter('obj', dsl.MuranoObjectParameter(decorate=False))
@specs.parameter('property_name', yaqltypes.Keyword())
@specs.name('#operator_.')
def obj_attribution(context, obj, property_name):
    return obj.get_property(property_name, context)


@specs.parameter('cls', dsl_types.MuranoTypeReference)
@specs.parameter('property_name', yaqltypes.Keyword())
@specs.name('#operator_.')
def obj_attribution_static(context, cls, property_name):
    return cls.type.get_property(property_name, context)


@specs.parameter('receiver', dsl.MuranoObjectParameter(decorate=False))
@specs.parameter('expr', yaqltypes.Lambda(method=True))
@specs.inject('operator', yaqltypes.Super(with_context=True))
@specs.name('#operator_.')
def op_dot(context, receiver, expr, operator):
    executor = helpers.get_executor()
    type_context = executor.context_manager.create_type_context(receiver.type)
    obj_context = executor.context_manager.create_object_context(receiver)
    ctx2 = helpers.link_contexts(
        helpers.link_contexts(context, type_context),
        obj_context)
    return operator(ctx2, receiver, expr)


@specs.parameter('receiver', dsl_types.MuranoTypeReference)
@specs.parameter('expr', yaqltypes.Lambda(method=True))
@specs.inject('operator', yaqltypes.Super(with_context=True))
@specs.name('#operator_.')
def op_dot_static(context, receiver, expr, operator):
    executor = helpers.get_executor()
    type_context = executor.context_manager.create_type_context(
        receiver.type)
    ctx2 = helpers.link_contexts(context, type_context)
    return operator(ctx2, receiver, expr)


@specs.parameter('prefix', yaqltypes.Keyword())
@specs.parameter('name', yaqltypes.Keyword())
@specs.name('#operator_:')
def ns_resolve(context, prefix, name):
    return helpers.get_class(prefix + ':' + name, context).get_reference()


@specs.parameter('name', yaqltypes.Keyword())
@specs.name('#unary_operator_:')
def ns_resolve_unary(context, name):
    return ns_resolve(context, '', name)


@specs.parameter('obj', dsl.MuranoObjectParameter(nullable=True))
@specs.parameter('type_', dsl.MuranoTypeParameter())
@specs.name('#operator_is')
def is_instance_of(obj, type_):
    if obj is None:
        return False
    return type_.type.is_compatible(obj)


def is_object(value):
    return isinstance(value, (dsl_types.MuranoObject,
                              dsl_types.MuranoTypeReference))


@specs.name('call')
@specs.parameter('name', yaqltypes.String())
@specs.parameter('args', yaqltypes.Sequence())
@specs.parameter('kwargs', utils.MappingType)
@specs.inject('op_dot', yaqltypes.Delegate('#operator_.', with_context=True))
@specs.inject('base', yaqltypes.Super(with_context=True))
def call_func(context, op_dot, base, name, args, kwargs,
              receiver=utils.NO_VALUE):
    if isinstance(receiver, (dsl_types.MuranoObject,
                             dsl_types.MuranoTypeReference)):
        kwargs = utils.filter_parameters_dict(kwargs)
        args += tuple(
            expressions.MappingRuleExpression(expressions.KeywordConstant(key),
                                              value)
            for key, value in six.iteritems(kwargs))
        function = expressions.Function(name, *args)
        return op_dot(context, receiver, function)
    else:
        return base(context, name, args, kwargs, receiver)


@specs.parameter('obj', dsl.MuranoObjectParameter(decorate=False))
@specs.parameter('serialization_type', yaqltypes.String())
@specs.parameter('ignore_upcasts', bool)
def dump(obj, serialization_type=dsl_types.DumpTypes.Serializable,
         ignore_upcasts=True):
    if serialization_type not in dsl_types.DumpTypes.All:
        raise ValueError('Invalid Serialization Type')
    executor = helpers.get_executor()
    if ignore_upcasts:
        obj = obj.real_this
    return serializer.serialize(obj, executor, serialization_type)


def register(context, runtime_version):
    context.register_function(cast)
    context.register_function(new)
    context.register_function(new_from_dict)
    context.register_function(new_from_model)
    context.register_function(id_)
    context.register_function(super_)
    context.register_function(psuper)
    context.register_function(require)
    context.register_function(find)
    context.register_function(sleep_)
    context.register_function(typeinfo)
    context.register_function(typeinfo_for_class)
    context.register_function(name)
    context.register_function(obj_attribution)
    context.register_function(obj_attribution_static)
    context.register_function(op_dot)
    context.register_function(op_dot_static)
    context.register_function(ns_resolve)
    context.register_function(ns_resolve_unary)
    reflection.register(context)
    context.register_function(is_instance_of)
    if runtime_version <= constants.RUNTIME_VERSION_1_3:
        context.register_function(type_legacy)
    else:
        context.register_function(type_)
    context.register_function(call_func)

    if runtime_version <= constants.RUNTIME_VERSION_1_1:
        context = context.create_child_context()
        for t in ('id', 'cast', 'super', 'psuper', 'type'):
            for spec in utils.to_extension_method(t, context):
                context.register_function(spec)

    context.register_function(type_from_name)
    context.register_function(is_object)
    context.register_function(dump)
    return context



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

import cloudify_rest_client
import cloudify_rest_client.exceptions as cloudify_exceptions
from murano.dsl import dsl
from oslo_config import cfg as config
from yaql.language import specs
from yaql.language import yaqltypes

import cfg


CONF = config.CONF
archive_upload_lock = threading.Lock()


class CloudifyClient(object):
    @specs.parameter('app', dsl.MuranoObjectParameter('io.murano.Application'))
    def __init__(self, app):
        cloudify_manager = self.CONF.cloudify_manager
        self._client = cloudify_rest_client.CloudifyClient(cloudify_manager)
        self._blueprint_id = '{0}-{1}'.format(app.type.name, app.type.version)
        self._deployment_id = app.id
        self._application_package = app.package

    @specs.parameter('entry_point', yaqltypes.String())
    def publish_blueprint(self, entry_point):
        global archive_upload_lock

        if self._check_blueprint_exists():
            return
        path = self._application_package.get_resource(entry_point)
        with archive_upload_lock:
            try:
                self._client.blueprints.upload(
                    path, self._blueprint_id)
            except cloudify_exceptions.CloudifyClientError as e:
                if e.status_code != 409:
                    raise

    def _check_blueprint_exists(self):
        try:
            self._client.blueprints.get(self._blueprint_id)
            return True
        except cloudify_exceptions.CloudifyClientError as e:
            if e.status_code == 404:
                return False
            raise

    @specs.parameter('parameters', dict)
    def create_deployment(self, parameters=None):
        self._client.deployments.create(
            self._blueprint_id, self._deployment_id, parameters)

    def delete_deployment(self):
        self._client.deployments.delete(self._deployment_id)

    def wait_deployment_ready(self):
        while True:
            executions = self._client.executions.list(self._deployment_id)
            if any(t.status in ('pending', 'started') for t in executions):
                time.sleep(3)
            else:
                deployment = self._client.deployments.get(self._deployment_id)
                return deployment.outputs

    @specs.parameter('name', yaqltypes.String())
    @specs.parameter('parameters', dict)
    def execute_workflow(self, name, parameters=None):
        self._client.executions.start(self._deployment_id, name, parameters)

    @classmethod
    def init_plugin(cls):
        cls.CONF = cfg.init_config(CONF)



#    Copyright (c) 2015 Telefonica I+D
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import tempfile

import mock
from oslo_serialization import base64
import yaml as yamllib

from murano.dsl import murano_object
from murano.dsl import murano_type
from murano.dsl import object_store
from murano.engine.system import agent
from murano.engine.system import resource_manager
from murano.tests.unit import base


class TestExecutionPlan(base.MuranoTestCase):
    def setUp(self):
        super(TestExecutionPlan, self).setUp()
        if hasattr(yamllib, 'CSafeLoader'):
            self.yaml_loader = yamllib.CSafeLoader
        else:
            self.yaml_loader = yamllib.SafeLoader

        self.mock_murano_class = mock.Mock(spec=murano_type.MuranoClass)
        self.mock_murano_class.name = 'io.murano.system.Agent'
        self.mock_murano_class.declared_parents = []
        self.mock_object_store = mock.Mock(spec=object_store.ObjectStore)

        object_interface = mock.Mock(spec=murano_object.MuranoObject)
        object_interface.id = '1234'
        object_interface.find_owner = lambda *args, **kwargs: object_interface

        self.agent = agent.Agent(object_interface)
        self.resources = mock.Mock(spec=resource_manager.ResourceManager)
        self.resources.string.return_value = 'text'
        self.uuids = ['ID1', 'ID2', 'ID3', 'ID4']
        self.mock_uuid = self._stub_uuid(self.uuids)
        self.addCleanup(mock.patch.stopall)

    def _read(self, path):
        execution_plan_dir = os.path.abspath(
            os.path.join(__file__, '../execution_plans/')
        )
        with open(execution_plan_dir + "/" + path) as file:
            return file.read()

    def test_execution_plan_v2_application_type(self):
        template = yamllib.load(
            self._read('application.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_application(), template)

    def test_execution_plan_v2_chef_type(self):
        template = yamllib.load(
            self._read('chef.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_chef(), template)

    def test_execution_plan_v2_telnet_application(self):
        template = yamllib.load(
            self._read('DeployTelnet.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_telnet_application(), template)

    def test_execution_plan_v2_tomcat_application(self):
        template = yamllib.load(
            self._read('DeployTomcat.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)

    def test_execution_plan_v2_app_without_files(self):
        template = yamllib.load(
            self._read('application_without_files.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_app_without_files(), template)

    def test_execution_plan_v2_app_with_file_in_template(self):
        template = yamllib.load(
            self._read('template_with_files.template'),
            Loader=self.yaml_loader)
        template = self.agent.build_execution_plan(template, self.resources)
        self.assertEqual(self._get_app_with_files_in_template(), template)

    def _get_application(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
                self.uuids[2]: {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'installer'
                },
                self.uuids[3]: {
                    'Body': 'dGV4dA==\n',
                    'BodyType': 'Base64',
                    'Name': 'common.sh'
                }
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Name': 'Deploy Tomcat',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [
                        self.uuids[2],
                        self.uuids[3]
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_app_with_files_in_template(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
                'updateScript': {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'updateScript'
                },
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Name': 'Deploy Tomcat',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [
                        'updateScript'
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_app_without_files(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTomcat.sh'
                },
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Name': 'Deploy Tomcat',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_chef(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Name': 'tomcat.git',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/tomcat.git'
                },
                self.uuids[2]: {
                    'Name': 'java',
                    'Type': 'Downloadable',
                    'URL': 'https://github.com/java.git'
                },

            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Name': 'Deploy Chef',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': 'cookbook/recipe',
                    'Files': [
                        self.uuids[1],
                        self.uuids[2]
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Chef',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _get_telnet_application(self):
        return {
            'Action': 'Execute',
            'Body': 'return deploy(args.appName).stdout\n',
            'Files': {
                self.uuids[1]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'deployTelnet.sh'
                },
                self.uuids[2]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'installer.sh'
                },
                self.uuids[3]: {
                    'Body': 'text',
                    'BodyType': 'Text',
                    'Name': 'common.sh'
                }
            },
            'FormatVersion': '2.0.0',
            'ID': self.uuids[0],
            'Name': 'Deploy Telnet',
            'Parameters': {
                'appName': '$appName'
            },
            'Scripts': {
                'deploy': {
                    'EntryPoint': self.uuids[1],
                    'Files': [
                        self.uuids[2],
                        self.uuids[3]
                    ],
                    'Options': {
                        'captureStderr': True,
                        'captureStdout': True
                    },
                    'Type': 'Application',
                    'Version': '1.0.0'
                }
            },
            'Version': '1.0.0'
        }

    def _stub_uuid(self, values=None):
        class FakeUUID(object):
            def __init__(self, v):
                self.hex = v

        if values is None:
            values = []
        mock_uuid4 = mock.patch('uuid.uuid4').start()
        mock_uuid4.side_effect = [FakeUUID(v) for v in values]
        return mock_uuid4

    @mock.patch('murano.engine.system.resource_manager.ResourceManager'
                '._get_package')
    def test_file_line_endings(self, _get_package):
        class FakeResources(object):
            """Class with only string() method from ResourceManager class"""
            @staticmethod
            def string(name, owner=None, binary=False):
                return resource_manager.ResourceManager.string(
                    receiver=None, name=name, owner=owner, binary=binary)

        # make path equal to provided name inside resources.string()
        package = mock.Mock()
        package.get_resource.side_effect = lambda m: m
        _get_package.return_value = package

        text = b"First line\nSecond line\rThird line\r\nFourth line"
        modified_text = u"First line\nSecond line\nThird line\nFourth line"
        encoded_text = base64.encode_as_text(text) + "\n"
        resources = FakeResources()

        with tempfile.NamedTemporaryFile() as script_file:
            script_file.write(text)
            script_file.file.flush()
            os.fsync(script_file.file.fileno())

            # check that data has been written correctly
            script_file.seek(0)
            file_data = script_file.read()
            self.assertEqual(text, file_data)

            # check resources.string() output
            # text file
            result = resources.string(script_file.name)
            self.assertEqual(modified_text, result)
            # binary file
            result = resources.string(script_file.name, binary=True)
            self.assertEqual(text, result)

            # check _get_body() output
            filename = os.path.basename(script_file.name)
            folder = os.path.dirname(script_file.name)
            # text file
            body = self.agent._get_body(filename, resources, folder)
            self.assertEqual(modified_text, body)
            # binary file
            filename = '<{0}>'.format(filename)
            body = self.agent._get_body(filename, resources, folder)
            self.assertEqual(encoded_text, body)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils

from murano.api.v1 import environments
from murano.api.v1 import sessions
import murano.tests.unit.api.base as tb


class TestSessionsApi(tb.ControllerTest, tb.MuranoApiTestCase):
    def setUp(self):
        super(TestSessionsApi, self).setUp()
        self.environments_controller = environments.Controller()
        self.sessions_controller = sessions.Controller()
        self.fixture = self.useFixture(config_fixture.Config())
        self.fixture.conf(args=[])

    def test_cant_deploy_from_another_tenant(self):
        """Test to prevent deployment under another tenant user's creds

        If user from one tenant uses session id and environment id
        of user from another tenant - he is not able to deploy
        the environment.

        Bug: #1382026
        """
        CREDENTIALS_1 = {'tenant': 'test_tenant_1', 'user': 'test_user_1'}
        CREDENTIALS_2 = {'tenant': 'test_tenant_2', 'user': 'test_user_2'}

        self._set_policy_rules(
            {'create_environment': '@'}
        )
        self.expect_policy_check('create_environment')

        # Create environment for user #1
        request = self._post(
            '/environments',
            jsonutils.dump_as_bytes({'name': 'test_environment_1'}),
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)
        self.assertEqual(CREDENTIALS_1['tenant'],
                         response_body['tenant_id'])

        ENVIRONMENT_ID = response_body['id']

        # Create session of user #1
        request = self._post(
            '/environments/{environment_id}/configure'
            .format(environment_id=ENVIRONMENT_ID),
            b'',
            **CREDENTIALS_1
        )
        response_body = jsonutils.loads(request.get_response(self.api).body)

        SESSION_ID = response_body['id']

        # Deploy the environment using environment id and session id of user #1
        # by user #2
        request = self._post(
            '/environments/{environment_id}/sessions/'
            '{session_id}/deploy'
            .format(environment_id=ENVIRONMENT_ID, session_id=SESSION_ID),
            b'',
            **CREDENTIALS_2
        )
        response = request.get_response(self.api)

        # Should be forbidden!
        self.assertEqual(403, response.status_code)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import io
import os
import zipfile

import six


class PackageType(object):
    Library = 'Library'
    Application = 'Application'
    ALL = [Library, Application]


@six.add_metaclass(abc.ABCMeta)
class Package(object):
    def __init__(self, format_name, runtime_version, source_directory):
        self._source_directory = source_directory
        self._format_name = format_name
        self._runtime_version = runtime_version
        self._blob_cache = None

    @property
    def format_name(self):
        return self._format_name

    @abc.abstractproperty
    def full_name(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def version(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def classes(self):
        raise NotImplementedError()

    @property
    def runtime_version(self):
        return self._runtime_version

    @abc.abstractproperty
    def requirements(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def package_type(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def display_name(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def description(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def author(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def supplier(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def tags(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def logo(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def supplier_logo(self):
        raise NotImplementedError()

    @property
    def blob(self):
        if not self._blob_cache:
            self._blob_cache = _pack_dir(self._source_directory)
        return self._blob_cache

    @abc.abstractmethod
    def get_class(self, name):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_resource(self, name):
        raise NotImplementedError()

    @abc.abstractproperty
    def ui(self):
        raise NotImplementedError()

    @abc.abstractproperty
    def meta(self):
        raise NotImplementedError()


def _zip_dir(path, zip_file):
    for root, _, files in os.walk(path):
        for f in files:
            abs_path = os.path.join(root, f)
            relative_path = os.path.relpath(abs_path, path)
            zip_file.write(abs_path, relative_path)


def _pack_dir(source_directory):
    blob = io.BytesIO()
    zip_file = zipfile.ZipFile(blob, mode='w')
    _zip_dir(source_directory, zip_file)
    zip_file.close()

    return blob.getvalue()



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from murano.dsl import dsl
from murano.dsl import exceptions
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestContracts(test_case.DslTestCase):
    def setUp(self):
        super(TestContracts, self).setUp()
        self._runner = self.new_runner(
            om.Object(
                'ContractExamples',
                ordinaryProperty='PROPERTY',
                sampleClass=om.Object(
                    'SampleClass1',
                    stringProperty='string1',
                    classProperty=om.Object(
                        'SampleClass2',
                        class2Property='string2'))))

    def test_string_contract(self):
        result = self._runner.testStringContract('qwerty')
        self.assertIsInstance(result, six.string_types)
        self.assertEqual('qwerty', result)

    def test_string_from_number_contract(self):
        result = self._runner.testStringContract(123)
        self.assertIsInstance(result, six.string_types)
        self.assertEqual('123', result)

    def test_string_null_contract(self):
        self.assertIsNone(self._runner.testStringContract(None))

    def test_int_contract(self):
        result = self._runner.testIntContract(123)
        self.assertIsInstance(result, int)
        self.assertEqual(123, result)

    def test_int_from_string_contract(self):
        result = self._runner.testIntContract('456')
        self.assertIsInstance(result, int)
        self.assertEqual(456, result)

    def test_int_from_string_contract_failure(self):
        self.assertRaises(exceptions.ContractViolationException,
                          self._runner.testIntContract, 'nan')

    def test_int_null_contract(self):
        self.assertIsNone(self._runner.testIntContract(None))

    def test_bool_contract(self):
        result = self._runner.testBoolContract(True)
        self.assertIsInstance(result, bool)
        self.assertTrue(result)

        result = self._runner.testBoolContract(False)
        self.assertIsInstance(result, bool)
        self.assertFalse(result)

    def test_bool_from_int_contract(self):
        result = self._runner.testBoolContract(10)
        self.assertIsInstance(result, bool)
        self.assertTrue(result)

        result = self._runner.testBoolContract(0)
        self.assertIsInstance(result, bool)
        self.assertFalse(result)

    def test_bool_from_string_contract(self):
        result = self._runner.testBoolContract('something')
        self.assertIsInstance(result, bool)
        self.assertTrue(result)

        result = self._runner.testBoolContract('')
        self.assertIsInstance(result, bool)
        self.assertFalse(result)

    def test_bool_null_contract(self):
        self.assertIsNone(self._runner.testIntContract(None))

    def test_class_contract(self):
        arg = om.Object('SampleClass2', class2Property='qwerty')
        result = self._runner.testClassContract(arg)
        self.assertIsInstance(result, dsl.MuranoObjectInterface)

    def test_class_contract_by_ref(self):
        arg = om.Object('SampleClass2', class2Property='qwerty')
        result = self._runner.testClassContract(arg)
        self.assertNotEqual(arg.id, result.id)

    def test_class_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testClassContract, ['invalid type'])

    def test_class_contract_by_ref_failure(self):
        self.assertRaises(
            exceptions.NoObjectFoundError,
            self._runner.testClassContract, 'NoSuchIdExists')

    def test_class_contract_from_dict(self):
        self.assertEqual(
            'SampleClass2',
            self._runner.testClassContract({
                'class2Property': 'str'}).type.name)

    def test_class_from_id_contract(self):
        object_id = self._runner.root.get_property('sampleClass').object_id
        result = self._runner.testClassFromIdContract(object_id)
        self.assertIsInstance(result, dsl.MuranoObjectInterface)
        self.assertEqual(object_id, result.id)

    def test_template_contract(self):
        arg = om.Object('CreatedClass2', property1='qwerty', property2=123)
        result = self._runner.testTemplateContract(arg)
        self.assertIsInstance(result, dict)
        self.assertItemsEqual(['?', 'property1', 'property2'], result.keys())

    def test_template_property_contract(self):
        template = {
            'foo': 123
        }
        self.new_runner(
            om.Object('ContractExamples', templateProperty=template))

    def test_template_contract_fail_on_type(self):
        arg = om.Object('SampleClass2', class2Property='qwerty')
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testTemplateContract, arg)

    def test_template_contract_with_property_exclusion(self):
        arg = om.Object('CreatedClass2', property1='qwerty',
                        property2='INVALID')
        result = self._runner.testTemplateContractExcludeProperty(arg)
        self.assertIsInstance(result, dict)
        self.assertItemsEqual(['?', 'property1'], result.keys())

    def test_template_contract_with_property_exclusion_from_mpl(self):
        result = self._runner.testTemplateContractExcludePropertyFromMpl()
        self.assertIsInstance(result, dict)
        self.assertItemsEqual(['?', 'property1'], result.keys())

    def test_check_contract(self):
        arg = om.Object('SampleClass2', class2Property='qwerty')
        self.assertIsNone(self._runner.testCheckContract(arg, 100))

    def test_check_contract_failure(self):
        invalid_arg = om.Object('SampleClass2', class2Property='not qwerty')
        self.assertRaises(exceptions.ContractViolationException,
                          self._runner.testCheckContract, invalid_arg, 100)

    def test_owned_contract(self):
        arg1 = self._runner.root.get_property('sampleClass')
        arg2 = arg1.get_property('classProperty')
        self.assertIsNone(self._runner.testOwnedContract(arg1, arg2))

    def test_owned_contract_on_null(self):
        self.assertIsNone(self._runner.testOwnedContract(None, None))

    def test_owned_contract_failure(self):
        arg1 = self._runner.root.get_property('sampleClass')
        arg2 = arg1.get_property('classProperty')
        invalid_arg2 = om.Object('SampleClass2', class2Property='string2')
        invalid_arg1 = om.Object(
            'SampleClass1',
            stringProperty='string1',
            classProperty=invalid_arg2)

        self.assertRaises(exceptions.ContractViolationException,
                          self._runner.testOwnedContract, invalid_arg1, arg2)
        self.assertRaises(exceptions.ContractViolationException,
                          self._runner.testOwnedContract, invalid_arg2, arg1)

    def test_not_owned_contract(self):
        arg2 = om.Object('SampleClass2', class2Property='string2')
        arg1 = om.Object(
            'SampleClass1',
            stringProperty='string1',
            classProperty=arg2)
        self.assertIsNone(self._runner.testNotOwnedContract(arg1, arg2))

    def test_not_owned_contract_on_null(self):
        self.assertIsNone(self._runner.testNotOwnedContract(None, None))

    def test_not_owned_contract_failure(self):
        invalid_arg1 = self._runner.root.get_property('sampleClass')
        invalid_arg2 = invalid_arg1.get_property('classProperty')
        arg2 = om.Object('SampleClass2', class2Property='string2')
        arg1 = om.Object(
            'SampleClass1',
            stringProperty='string1',
            classProperty=arg2)

        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testNotOwnedContract, invalid_arg1, arg2)
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testNotOwnedContract, invalid_arg2, arg1)

    def test_scalar_contract(self):
        self.assertEqual('fixed', self._runner.testScalarContract(
            'fixed', 456, True))

    def test_scalar_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testScalarContract,
            'wrong', 456, True)

        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testScalarContract,
            'fixed', 123, True)

        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testScalarContract,
            'fixed', 456, False)

    def test_list_contract(self):
        self.assertEqual([3, 2, 1], self._runner.testListContract(
            ['3', 2, '1']))

    def test_list_contract_from_scalar(self):
        self.assertEqual([99], self._runner.testListContract('99'))

    def test_list_contract_from_null(self):
        self.assertEqual([], self._runner.testListContract(None))

    def test_list_with_min_length_contract(self):
        self.assertEqual(
            [1, 2, 3],
            self._runner.testListWithMinLengthContract([1, 2, 3]))
        self.assertEqual(
            [1, 2, 3, 4],
            self._runner.testListWithMinLengthContract([1, 2, 3, 4]))

    def test_list_with_min_length_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testListWithMinLengthContract, None)
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testListWithMinLengthContract, [1, 2])

    def test_list_with_min_max_length_contract(self):
        self.assertEqual(
            [1, 2],
            self._runner.testListWithMinMaxLengthContract([1, 2]))
        self.assertEqual(
            [1, 2, 3, 4],
            self._runner.testListWithMinMaxLengthContract([1, 2, 3, 4]))

    def test_list_with_min_max_length_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testListWithMinMaxLengthContract, [1])
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testListWithMinMaxLengthContract, [1, 2, 3, 4, 5])

    def test_dict_contract(self):
        self.assertEqual(
            {'A': '123', 'B': 456},
            self._runner.testDictContract({'A': '123', 'B': '456'}))
        self.assertEqual(
            {'A': '123', 'B': 456},
            self._runner.testDictContract({'A': '123', 'B': '456', 'C': 'qq'}))
        self.assertEqual(
            {'A': '123', 'B': None},
            self._runner.testDictContract({'A': '123'}))

    def test_dict_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testDictContract, 'str')

    def test_dict_expressions_contract(self):
        self.assertEqual(
            {321: 'qwerty', 99: 'val', 'B': 456},
            self._runner.testDictExprContract({
                '321': 'qwerty', '99': 'val', 'B': 456}))

    def test_dict_expressions_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testDictExprContract,
            {'321': 'qwerty', 'str': 'val', 'B': 456})

    def test_invalid_dict_expr_contract(self):
        self.assertRaises(
            exceptions.DslContractSyntaxError,
            self._runner.testDictMultiExprContract,
            {'321': 'qwerty', 'str': 'val', 'B': 456})

    def test_not_null_contract(self):
        self.assertEqual('value', self._runner.testNotNullContract('value'))

    def test_not_null_contract_failure(self):
        self.assertRaises(
            exceptions.ContractViolationException,
            self._runner.testNotNullContract, None)

    def test_default(self):
        self.assertEqual('value', self._runner.testDefault('value'))
        self.assertEqual('DEFAULT', self._runner.testDefault())

    def test_default_expression(self):
        self.assertEqual('PROPERTY', self._runner.testDefaultExpression())
        self.assertEqual('value', self._runner.testDefaultExpression('value'))


class TestContractsTransform(test_case.DslTestCase):
    def setUp(self):
        super(TestContractsTransform, self).setUp()
        self._runner = self.new_runner(om.Object('TestIteratorsTransform'))

    def test_property(self):
        self.assertEqual('3', self._runner.testProperties())

    def test_argument(self):
        self.assertEqual('3', self._runner.testArgs())
        self.assertEqual('2', self._runner.testUntypedArgs())
        self.assertEqual('6', self._runner.testNotTypedListArgs())
        self.assertEqual('6', self._runner.testTypedList())
        self.assertEqual(2, self._runner.testListDict())



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import operator
import weakref

import six

from murano.dsl import dsl_types
from murano.dsl import helpers


class MetaProvider(object):
    @abc.abstractmethod
    def get_meta(self, context):
        raise NotImplementedError()


class MetaData(MetaProvider):
    def __init__(self, definition, target, declaring_type):
        declaring_type = weakref.proxy(declaring_type)
        definition = helpers.list_value(definition)
        factories = []
        used_types = set()
        for d in definition:
            if isinstance(d, dict):
                if len(d) != 1:
                    raise ValueError('Invalid Meta format')
                name = next(iter(d.keys()))
                props = d[name] or {}
            else:
                name = d
                props = {}
            type_obj = helpers.resolve_type(name, declaring_type)
            if type_obj.usage != dsl_types.ClassUsages.Meta:
                raise ValueError('Only Meta classes can be attached')
            if target not in type_obj.targets:
                raise ValueError(
                    u'Meta class {} is not applicable here'.format(
                        type_obj.name))
            if type_obj in used_types and (
                    type_obj.cardinality != dsl_types.MetaCardinality.Many):
                raise ValueError('Cannot attach several Meta instances '
                                 'with cardinality One')

            used_types.add(type_obj)

            def factory_maker(template):
                def instantiate(context):
                    obj = helpers.get_object_store().load(
                        template, owner=None,
                        context=context, scope_type=declaring_type)
                    obj.declaring_type = declaring_type
                    return obj
                return instantiate

            factories.append(factory_maker({type_obj: props}))
        self._meta_factories = factories
        self._meta = None

    def get_meta(self, context):
        if self._meta is None:
            self._meta = list(map(lambda x: x(context), self._meta_factories))
        return self._meta


def merge_providers(initial_class, producer, context):
    def merger(cls_list, skip_list):
        result = set()
        all_meta = []
        for cls in cls_list:
            cls_skip_list = skip_list.copy()
            provider = producer(cls)
            meta = [] if provider is None else provider.get_meta(context)
            for item in meta:
                cardinality = item.type.cardinality
                inherited = item.type.inherited
                if cls != initial_class and (
                        not inherited or item.type in skip_list):
                    continue
                if cardinality == dsl_types.MetaCardinality.One:
                    cls_skip_list.add(item.type)
                all_meta.append((cls, item))
            all_meta.extend(merger(cls.parents, cls_skip_list))
        meta_types = {}
        for cls, item in all_meta:
            entry = meta_types.get(item.type)
            if entry is not None:
                if entry != cls:
                    raise ValueError(
                        u'Found more than one instance of Meta {} '
                        u'with Cardinality One'.format(item.type.name))
                else:
                    continue

            if item.type.cardinality == dsl_types.MetaCardinality.One:
                meta_types[item.type] = cls
            result.add((cls, item))
        return result

    meta = merger([initial_class], set())
    return list(six.moves.map(operator.itemgetter(1), meta))


def aggregate_meta(provider, context, group_by_name=True):
    key_func = lambda m: m.type.name if group_by_name else m.type
    meta = provider.get_meta(context)
    result = {}
    for item in meta:
        if item.type.cardinality == dsl_types.MetaCardinality.One:
            result[key_func(item)] = item
        else:
            result.setdefault(key_func(item), []).append(item)
    return result



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import congressclient.v1.client as cclient
import keystoneclient
from keystoneclient.v2_0 import client as ksclient
from tempest import config

import murano.tests.functional.common.utils as common_utils

CONF = config.CONF


class TempestDeployTestMixin(common_utils.DeployTestMixin):
    """Overrides methods to use tempest configuration."""

    @staticmethod
    @common_utils.memoize
    def keystone_client():
        return ksclient.Client(username=CONF.auth.admin_username,
                               password=CONF.auth.admin_password,
                               tenant_name=CONF.auth.admin_project_name,
                               auth_url=CONF.identity.uri)

    @staticmethod
    @common_utils.memoize
    def congress_client():
        auth = keystoneclient.auth.identity.v2.Password(
            auth_url=CONF.identity.uri,
            username=CONF.auth.admin_username,
            password=CONF.auth.admin_password,
            tenant_name=CONF.auth.admin_project_name)
        session = keystoneclient.session.Session(auth=auth)
        return cclient.Client(session=session,
                              service_type='policy')



# Copyright (c) 2013 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from murano.engine.system import agent
from murano.engine.system import agent_listener
from murano.engine.system import garbage_collector
from murano.engine.system import heat_stack
from murano.engine.system import instance_reporter
from murano.engine.system import logger
from murano.engine.system import net_explorer
from murano.engine.system import resource_manager
from murano.engine.system import status_reporter
from murano.engine.system import test_fixture
from murano.engine.system import workflowclient


def register(package):
    package.register_class(agent.Agent)
    package.register_class(agent_listener.AgentListener)
    package.register_class(heat_stack.HeatStack)
    package.register_class(resource_manager.ResourceManager)
    package.register_class(instance_reporter.InstanceReportNotifier)
    package.register_class(status_reporter.StatusReporter)
    package.register_class(net_explorer.NetworkExplorer)
    package.register_class(logger.Logger)
    package.register_class(test_fixture.TestFixture)
    package.register_class(workflowclient.MistralClient)
    package.register_class(garbage_collector.GarbageCollector)



#    Copyright (c) 2014 Mirantis, Inc.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import cgi
import os
import tempfile

import jsonschema
from keystoneclient import exceptions as keystone_ex
from keystoneclient import service_catalog
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_log import versionutils
import six
from webob import exc

import murano.api.v1
from murano.api.v1 import validation_schemas
from murano.common import exceptions
from murano.common import policy
import murano.common.utils as murano_utils
from murano.common import wsgi
from murano.db.catalog import api as db_api
from murano.common.i18n import _, _LW
from murano.packages import exceptions as pkg_exc
from murano.packages import load_utils
from muranoclient.glance import client as glare_client


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

SUPPORTED_PARAMS = murano.api.v1.SUPPORTED_PARAMS
LIST_PARAMS = murano.api.v1.LIST_PARAMS
ORDER_VALUES = murano.api.v1.ORDER_VALUES
PKG_PARAMS_MAP = murano.api.v1.PKG_PARAMS_MAP
OPERATOR_VALUES = murano.api.v1.OPERATOR_VALUES


def _check_content_type(req, content_type):
    try:
        req.get_content_type((content_type,))
    except exceptions.UnsupportedContentType:
        msg = _("Content-Type must be '{type}'").format(type=content_type)
        LOG.error(msg)
        raise exc.HTTPBadRequest(explanation=msg)


def _get_filters(query_params):
    filters = {}
    for param_pair in query_params:
        k, v = param_pair
        if k not in SUPPORTED_PARAMS:
            LOG.warning(_LW("Search by parameter '{name}' "
                            "is not supported. Skipping it.").format(name=k))
            continue

        if k in LIST_PARAMS:
            if v.startswith('in:') and k in OPERATOR_VALUES:
                in_value = v[len('in:'):]
                try:
                    filters[k] = murano_utils.split_for_quotes(in_value)
                except ValueError as err:
                    LOG.warning(_LW("Search by parameter '{name}' "
                                    "caused an {message} error."
                                    "Skipping it.").format(name=k,
                                                           message=err))
            else:
                filters.setdefault(k, []).append(v)
        else:
            filters[k] = v
        order_by = filters.get('order_by', [])
        for i in order_by[:]:
            if ORDER_VALUES and i not in ORDER_VALUES:
                filters['order_by'].remove(i)
                LOG.warning(_LW("Value of 'order_by' parameter is not valid. "
                                "Allowed values are: {values}. Skipping it.")
                            .format(values=", ".join(ORDER_VALUES)))
    return filters


def _validate_body(body):
    """Check multipart/form-data has two parts

    Check multipart/form-data has two parts: text (which is json string and
    should parsed into dictionary in serializer) and file, which stores as
    cgi.FieldStorage instance. Also validate file size doesn't exceed
    the limit: seek to the end of the file, get the position of EOF and
    reset the file position to the beginning
    """
    def check_file_size(f):
        mb_limit = CONF.murano.package_size_limit
        pkg_size_limit = mb_limit * 1024 * 1024
        f.seek(0, 2)
        size = f.tell()
        f.seek(0)
        if size > pkg_size_limit:
            raise exc.HTTPBadRequest(explanation=_(
                'Uploading file is too large. '
                'The limit is {0} Mb').format(mb_limit))

    if len(body) > 2:
        msg = _("'multipart/form-data' request body should contain 1 or 2 "
                "parts: json string and zip archive. Current body consists "
                "of {amount} part(s)").format(amount=len(body.keys()))
        LOG.error(msg)
        raise exc.HTTPBadRequest(explanation=msg)

    file_obj = None
    package_meta = None
    for part in body.values():
        if isinstance(part, cgi.FieldStorage):
            file_obj = part
            check_file_size(file_obj.file)

        if isinstance(part, dict):
            package_meta = part
    if file_obj is None:
        msg = _('There is no file package with application description')
        LOG.error(msg)
        raise exc.HTTPBadRequest(explanation=msg)
    return file_obj, package_meta


class Controller(object):
    """WSGI controller for application catalog resource in Murano v1 API."""

    def _validate_limit(self, value):
            if value is None:
                return
            try:
                value = int(value)
            except ValueError:
                msg = _("Limit param must be an integer")
                LOG.error(msg)
                raise exc.HTTPBadRequest(explanation=msg)

            if value <= 0:
                msg = _("Limit param must be positive")
                LOG.error(msg)
                raise exc.HTTPBadRequest(explanation=msg)

            return value

    def update(self, req, body, package_id):
        """List of allowed changes

        List of allowed changes:
            { "op": "add", "path": "/tags", "value": [ "foo", "bar" ] }
            { "op": "add", "path": "/categories", "value": [ "foo", "bar" ] }
            { "op": "remove", "path": "/tags" }
            { "op": "remove", "path": "/categories" }
            { "op": "replace", "path": "/tags", "value": ["foo", "bar"] }
            { "op": "replace", "path": "/is_public", "value": true }
            { "op": "replace", "path": "/description",
                                "value":"New description" }
            { "op": "replace", "path": "/name", "value": "New name" }
        """
        policy.check("modify_package", req.context, {'package_id': package_id})

        pkg_to_update = db_api.package_get(package_id, req.context)
        if pkg_to_update.is_public:
            policy.check("manage_public_package", req.context)

        _check_content_type(req, 'application/murano-packages-json-patch')
        if not isinstance(body, list):
            msg = _('Request body must be a JSON array of operation objects.')
            LOG.error(msg)
            raise exc.HTTPBadRequest(explanation=msg)
        for change in body:
            if 'is_public' in change['path']:
                if change['value'] is True and not pkg_to_update.is_public:
                    policy.check('publicize_package', req.context)
            if 'name' in change['path']:
                if len(change['value']) > 80:
                    msg = _('Package name should be 80 characters maximum')
                    LOG.error(msg)
                    raise exc.HTTPBadRequest(explanation=msg)
        package = db_api.package_update(package_id, body, req.context)
        return package.to_dict()

    def get(self, req, package_id):
        policy.check("get_package", req.context, {'package_id': package_id})

        package = db_api.package_get(package_id, req.context)
        return package.to_dict()

    def search(self, req):
        policy.check("get_package", req.context)
        manage_public = True
        try:
            policy.check("manage_public_package", req.context)
        except exc.HTTPForbidden:
            manage_public = False

        filters = _get_filters(req.GET.items())

        limit = self._validate_limit(filters.get('limit'))
        if limit is None:
            limit = CONF.murano.limit_param_default
        limit = min(CONF.murano.api_limit_max, limit)

        result = {}

        catalog = req.GET.pop('catalog', '').lower() == 'true'
        packages = db_api.package_search(
            filters, req.context, manage_public, limit, catalog=catalog)
        if len(packages) == limit:
            result['next_marker'] = packages[-1].id
        result['packages'] = [package.to_dict() for package in packages]
        return result

    def upload(self, req, body=None):
        """Upload new file archive

        Upload new file archive for the new package
        together with package metadata.
        """
        policy.check("upload_package", req.context)

        _check_content_type(req, 'multipart/form-data')
        file_obj, package_meta = _validate_body(body)
        if package_meta:
            try:
                jsonschema.validate(package_meta,
                                    validation_schemas.PKG_UPLOAD_SCHEMA)
            except jsonschema.ValidationError as e:
                msg = _("Package schema is not valid: {reason}").format(
                    reason=e)
                LOG.exception(msg)
                raise exc.HTTPBadRequest(explanation=msg)
        else:
            package_meta = {}

        if package_meta.get('is_public'):
            policy.check('publicize_package', req.context)

        with tempfile.NamedTemporaryFile(delete=False) as tempf:
            LOG.debug("Storing package archive in a temporary file")
            content = file_obj.file.read()
            if not content:
                msg = _("Uploading file can't be empty")
                LOG.error(msg)
                raise exc.HTTPBadRequest(explanation=msg)
            tempf.write(content)
            package_meta['archive'] = content
        try:
            with load_utils.load_from_file(
                    tempf.name, target_dir=None,
                    drop_dir=True) as pkg_to_upload:
                # extend dictionary for update db
                for k, v in six.iteritems(PKG_PARAMS_MAP):
                    if hasattr(pkg_to_upload, k):
                        package_meta[v] = getattr(pkg_to_upload, k)
                if len(package_meta['name']) > 80:
                    msg = _('Package name should be 80 characters maximum')
                    LOG.error(msg)
                    raise exc.HTTPBadRequest(explanation=msg)
                try:
                    package = db_api.package_upload(
                        package_meta, req.context.tenant)
                except db_exc.DBDuplicateEntry:
                    msg = _('Package with specified full '
                            'name is already registered')
                    LOG.exception(msg)
                    raise exc.HTTPConflict(msg)
                return package.to_dict()
        except pkg_exc.PackageLoadError as e:
            msg = _("Couldn't load package from file: {reason}").format(
                reason=e)
            LOG.exception(msg)
            raise exc.HTTPBadRequest(explanation=msg)
        finally:
            LOG.debug("Deleting package archive temporary file")
            os.remove(tempf.name)

    def get_ui(self, req, package_id):
        if CONF.engine.packages_service == 'murano':
            target = {'package_id': package_id}
            policy.check("get_package", req.context, target)

            package = db_api.package_get(package_id, req.context)
            return package.ui_definition
        else:
            g_client = self._get_glare_client(req)
            blob_data = g_client.artifacts.download_blob(package_id, 'archive')
            with tempfile.NamedTemporaryFile() as tempf:
                for chunk in blob_data:
                    tempf.write(chunk)
                tempf.file.flush()
                os.fsync(tempf.file.fileno())
                with load_utils.load_from_file(tempf.name, target_dir=None,
                                               drop_dir=True) as pkg:
                    return pkg.ui

    def get_logo(self, req, package_id):
        target = {'package_id': package_id}
        policy.check("get_package", req.context, target)

        package = db_api.package_get(package_id, req.context)
        return package.logo

    def get_supplier_logo(self, req, package_id):
        package = db_api.package_get(package_id, req.context)
        return package.supplier_logo

    def download(self, req, package_id):
        target = {'package_id': package_id}
        policy.check("download_package", req.context, target)

        package = db_api.package_get(package_id, req.context)
        return package.archive

    def delete(self, req, package_id):
        target = {'package_id': package_id}
        policy.check("delete_package", req.context, target)

        package = db_api.package_get(package_id, req.context)
        if package.is_public:
            policy.check("manage_public_package", req.context, target)
        db_api.package_delete(package_id, req.context)

    def get_category(self, req, category_id):
        policy.check("get_category", req.context)
        category = db_api.category_get(category_id, packages=True)
        return category.to_dict()

    @versionutils.deprecated(as_of=versionutils.deprecated.LIBERTY,
                             in_favor_of='categories.list()')
    def show_categories(self, req):
        policy.check("get_category", req.context)
        categories = db_api.categories_list()
        return {'categories': [category.name for category in categories]}

    def list_categories(self, req):
        """List all categories

        List all categories with pagination and sorting
           Acceptable filter params:
           :param sort_keys: an array of fields used to sort the list
           :param sort_dir: the direction of the sort ('asc' or 'desc')
           :param limit: the number of categories to list
           :param marker: the ID of the last item in the previous page
        """
        def _get_category_filters(req):
            query_params = {}
            valid_query_params = ['sort_keys', 'sort_dir', 'limit', 'marker']
            for key, value in req.GET.items():
                if key not in valid_query_params:
                    raise exc.HTTPBadRequest(
                        _('Bad value passed to filter. '
                          'Got {key}, expected:{valid}').format(
                            key=key, valid=', '.join(valid_query_params)))
                if key == 'sort_keys':
                    available_sort_keys = ['name', 'created',
                                           'updated', 'package_count', 'id']
                    value = [v.strip() for v in value.split(',')]
                    for sort_key in value:
                        if sort_key not in available_sort_keys:
                            raise exc.HTTPBadRequest(
                                explanation=_('Invalid sort key: {sort_key}. '
                                              'Must be one of the following: '
                                              '{available}').format(
                                    sort_key=sort_key,
                                    available=', '.join(available_sort_keys)))
                if key == 'sort_dir':
                    if value not in ['asc', 'desc']:
                        msg = _('Invalid sort direction: {0}').format(value)
                        raise exc.HTTPBadRequest(explanation=msg)
                query_params[key] = value
            return query_params

        policy.check("get_category", req.context)

        filters = _get_category_filters(req)

        marker = filters.get('marker')
        limit = self._validate_limit(filters.get('limit'))

        result = {}
        categories = db_api.categories_list(filters,
                                            limit=limit,
                                            marker=marker)
        if len(categories) == limit:
            result['next_marker'] = categories[-1].id

        result['categories'] = [category.to_dict() for category in categories]
        return result

    def add_category(self, req, body=None):
        policy.check("add_category", req.context)
        category_name = body.get('name')
        if not category_name:
            raise exc.HTTPBadRequest(
                explanation='Please, specify a name of the category to create')
        if len(category_name) > 80:
            msg = _('Category name should be 80 characters maximum')
            LOG.error(msg)
            raise exc.HTTPBadRequest(explanation=msg)
        try:
            category = db_api.category_add(category_name)
        except db_exc.DBDuplicateEntry:
            msg = _('Category with specified name is already exist')
            LOG.error(msg)
            raise exc.HTTPConflict(explanation=msg)
        return category.to_dict()

    def delete_category(self, req, category_id):
        target = {'category_id': category_id}
        policy.check("delete_category", req.context, target)
        category = db_api.category_get(category_id, packages=True)
        if category.packages:
            msg = _("It's impossible to delete categories assigned "
                    "to the package, uploaded to the catalog")
            raise exc.HTTPForbidden(explanation=msg)
        db_api.category_delete(category_id)

    def _get_glare_client(self, request):
        glare_settings = CONF.glare
        token = request.context.auth_token
        url = glare_settings.url
        if not url:
            url = self._get_glare_url(request)
        client = glare_client.Client(
            endpoint=url, token=token, insecure=glare_settings.insecure,
            key_file=glare_settings.key_file or None,
            ca_file=glare_settings.ca_file or None,
            cert_file=glare_settings.cert_file or None,
            type_name='murano',
            type_version=1)
        return client

    def _get_glare_url(self, request):
        sc = request.context.service_catalog
        token = request.context.auth_token
        try:
            return service_catalog.ServiceCatalogV2(
                {'serviceCatalog': sc}).url_for(
                service_type='artifact',
                endpoint_type=CONF.glare.endpoint_type,
                region_name=CONF.home_region)
        except keystone_ex.EndpointNotFound:
            return service_catalog.ServiceCatalogV3(
                token,
                {'catalog': sc}).url_for(
                    service_type='artifact',
                    endpoint_type=CONF.glare.endpoint_type,
                    region_name=CONF.home_region)


def create_resource():
    specific_content_types = {
        'get_ui': ['text/plain'],
        'download': ['application/octet-stream'],
        'get_logo': ['application/octet-stream'],
        'get_supplier_logo': ['application/octet-stream']}
    deserializer = wsgi.RequestDeserializer(
        specific_content_types=specific_content_types)
    return wsgi.Resource(Controller(), deserializer=deserializer)



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import testtools

from tempest import config
from tempest.lib import exceptions

from murano_tempest_tests.tests.api.application_catalog import base
from murano_tempest_tests import utils

CONF = config.CONF


class TestRepositoryNegativeNotFound(base.BaseApplicationCatalogTest):
    @classmethod
    def resource_setup(cls):
        if CONF.application_catalog.glare_backend:
            msg = ("Murano using GLARE backend. "
                   "Repository tests will be skipped.")
            raise cls.skipException(msg)
        super(TestRepositoryNegativeNotFound, cls).resource_setup()

    @testtools.testcase.attr('negative')
    def test_update_package_with_incorrect_id(self):

        post_body = [
            {
                "op": "add",
                "path": "/tags",
                "value": ["im a test"]
            }
        ]

        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.update_package,
                          utils.generate_uuid(), post_body)

    @testtools.testcase.attr('negative')
    def test_get_package_with_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.get_package,
                          utils.generate_uuid())

    @testtools.testcase.attr('negative')
    def test_delete_package_with_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.delete_package,
                          utils.generate_uuid())

    @testtools.testcase.attr('negative')
    def test_download_package_with_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.download_package,
                          utils.generate_uuid())

    @testtools.testcase.attr('negative')
    def test_get_ui_definition_with_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.get_ui_definition,
                          utils.generate_uuid())

    @testtools.testcase.attr('negative')
    def test_get_logo_with_incorrect_id(self):
        self.assertRaises(exceptions.NotFound,
                          self.application_catalog_client.get_logo,
                          utils.generate_uuid())


class TestRepositoryNegativeForbidden(base.BaseApplicationCatalogTest):

    # TODO(freerunner): I hope, that we can setup and cleanup resources
    # TODO(freerunner): dramatically better.
    @classmethod
    def resource_setup(cls):
        if CONF.application_catalog.glare_backend:
            msg = ("Murano using GLARE backend. "
                   "Repository tests will be skipped.")
            raise cls.skipException(msg)

        super(TestRepositoryNegativeForbidden, cls).resource_setup()

        application_name = utils.generate_name('package_test_upload')
        cls.abs_archive_path, dir_with_archive, archive_name = \
            utils.prepare_package(application_name)
        cls.package = cls.application_catalog_client.upload_package(
            application_name, archive_name, dir_with_archive,
            {"categories": [], "tags": [], 'is_public': False})
        cls.alt_client = cls.get_client_with_isolated_creds(
            type_of_creds='alt')

    @classmethod
    def resource_cleanup(cls):
        os.remove(cls.abs_archive_path)
        cls.application_catalog_client.delete_package(cls.package['id'])
        super(TestRepositoryNegativeForbidden, cls).resource_cleanup()

    @testtools.testcase.attr('negative')
    def test_update_package_from_another_tenant(self):
        post_body = [
            {
                "op": "add",
                "path": "/tags",
                "value": ["im a test"]
            }
        ]

        self.assertRaises(exceptions.Forbidden,
                          self.alt_client.update_package,
                          self.package['id'],
                          post_body)

    @testtools.testcase.attr('negative')
    def test_get_package_from_another_tenant(self):
        self.assertRaises(exceptions.Forbidden,
                          self.alt_client.get_package,
                          self.package['id'])

    @testtools.testcase.attr('negative')
    def test_delete_package_from_another_tenant(self):
        self.assertRaises(exceptions.Forbidden,
                          self.alt_client.delete_package,
                          self.package['id'])

    @testtools.testcase.attr('negative')
    def test_download_package_from_another_tenant(self):
        self.assertRaises(exceptions.Forbidden,
                          self.alt_client.download_package,
                          self.package['id'])

    @testtools.testcase.attr('negative')
    def test_get_ui_definition_from_another_tenant(self):
        self.assertRaises(exceptions.Forbidden,
                          self.alt_client.get_ui_definition,
                          self.package['id'])

    @testtools.testcase.attr('negative')
    def test_get_logo_from_another_tenant(self):
        self.assertRaises(exceptions.Forbidden,
                          self.alt_client.get_logo,
                          self.package['id'])



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This code is almost a complete copy of eventlet.corolocal with only
# the concept of current thread replaced with current session

import weakref

import six

from murano.dsl import helpers


# the entire purpose of this class is to store off the constructor
# arguments in a local variable without calling __init__ directly
class _localbase(object):
    __slots__ = '_local__args', '_local__sessions'

    def __new__(cls, *args, **kw):
        self = object.__new__(cls)
        object.__setattr__(self, '_local__args', (args, kw))
        object.__setattr__(
            self, '_local__sessions', weakref.WeakKeyDictionary())
        if (args or kw) and (cls.__init__ is object.__init__):
            raise TypeError('Initialization arguments are not supported')
        return self


def _patch(session_local):
    sessions_dict = object.__getattribute__(session_local, '_local__sessions')
    session = helpers.get_execution_session()
    localdict = sessions_dict.get(session)
    if localdict is None:
        # must be the first time we've seen this session, call __init__
        localdict = {}
        sessions_dict[session] = localdict
        cls = type(session_local)
        if cls.__init__ is not object.__init__:
            args, kw = object.__getattribute__(session_local, '_local__args')
            session_local.__init__(*args, **kw)
    object.__setattr__(session_local, '__dict__', localdict)


class _local(_localbase):
    def __getattribute__(self, attr):
        _patch(self)
        return object.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        _patch(self)
        return object.__setattr__(self, attr, value)

    def __delattr__(self, attr):
        _patch(self)
        return object.__delattr__(self, attr)


def session_local(cls):
    return type(cls.__name__, (cls, _local), {})


class SessionLocalDict(six.moves.UserDict, object):
    def __init__(self, **kwargs):
        self.__session_data = weakref.WeakKeyDictionary()
        self.__default = {}
        super(SessionLocalDict, self).__init__(**kwargs)

    @property
    def data(self):
        session = helpers.get_execution_session()
        if session is None:
            return self.__default
        return self.__session_data.setdefault(session, {})

    @data.setter
    def data(self, value):
        session = helpers.get_execution_session()
        if session is None:
            self.__default = value
        else:
            self.__session_data[session] = value


def execution_session_memoize(func):
    cache = SessionLocalDict()
    return helpers.get_memoize_func(func, cache)



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import six

from murano.dsl import context_manager
from murano.dsl import dsl
from murano.dsl import dsl_exception
from murano.dsl import dsl_types
from murano.dsl import executor
from murano.dsl import helpers
from murano.dsl import murano_object
from murano.dsl import serializer
from murano.dsl import yaql_integration
from murano.engine import execution_session
from murano.engine.system import yaql_functions
from murano.tests.unit.dsl.foundation import object_model


class TestContextManager(context_manager.ContextManager):
    def __init__(self, functions):
        self.__functions = functions

    def create_root_context(self, runtime_version):
        root_context = super(TestContextManager, self).create_root_context(
            runtime_version)
        context = helpers.link_contexts(
            root_context, yaql_functions.get_context(runtime_version))
        context = context.create_child_context()
        for name, func in six.iteritems(self.__functions):
            context.register_function(func, name)
        return context


class Runner(object):
    class DslObjectWrapper(object):
        def __init__(self, obj, runner):
            self._runner = runner
            if isinstance(obj, six.string_types + (dsl_types.MuranoType,)):
                pass
            elif isinstance(obj, (object_model.Object, object_model.Ref)):
                obj = obj.id
            elif isinstance(obj, murano_object.MuranoObject):
                obj = obj.object_id
            else:
                raise ValueError(
                    'obj must be object ID string, MuranoObject, MuranoType '
                    'or one of object_model helper classes (Object, Ref)')
            if isinstance(obj, six.string_types):
                self._receiver = runner.executor.object_store.get(obj)
            else:
                self._receiver = obj

            self._preserve_exception = False

        def __getattr__(self, item):
            def call(*args, **kwargs):
                return self._runner._execute(
                    item, self._receiver, *args, **kwargs)
            if item.startswith('test'):
                return call

    def __init__(self, model, package_loader, functions):
        if isinstance(model, six.string_types):
            model = object_model.Object(model)
        model = object_model.build_model(model)
        if 'Objects' not in model:
            model = {'Objects': model}

        self.executor = executor.MuranoDslExecutor(
            package_loader, TestContextManager(functions),
            execution_session.ExecutionSession())
        self._root = self.executor.load(model).object
        if 'ObjectsCopy' in model:
            self.executor.cleanup(model)

    def _execute(self, name, obj, *args, **kwargs):
        try:
            final_args = []
            final_kwargs = {}
            for arg in args:
                if isinstance(arg, object_model.Object):
                    arg = object_model.build_model(arg)
                final_args.append(arg)
            for name, arg in six.iteritems(kwargs):
                if isinstance(arg, object_model.Object):
                    arg = object_model.build_model(arg)
                final_kwargs[name] = arg
            cls = obj if isinstance(obj, dsl_types.MuranoType) else obj.type
            runtime_version = cls.package.runtime_version
            yaql_engine = yaql_integration.choose_yaql_engine(runtime_version)
            with helpers.with_object_store(self.executor.object_store):
                return dsl.to_mutable(cls.invoke(
                    name, obj, tuple(final_args), final_kwargs), yaql_engine)
        except dsl_exception.MuranoPlException as e:
            if not self.preserve_exception:
                original_exception = getattr(e, 'original_exception', None)
                if original_exception and not isinstance(
                        original_exception, dsl_exception.MuranoPlException):
                    exc_traceback = getattr(
                        e, 'original_traceback', None) or sys.exc_info()[2]
                    six.reraise(
                        type(original_exception),
                        original_exception,
                        exc_traceback)
            raise

    def __getattr__(self, item):
        if item.startswith('test'):
            return getattr(Runner.DslObjectWrapper(self._root, self), item)

    def on(self, obj):
        return Runner.DslObjectWrapper(obj, self)

    def on_class(self, class_name):
        cls = self.executor.package_loader.load_class_package(
            class_name, helpers.parse_version_spec(None)).find_class(
            class_name, False)
        return Runner.DslObjectWrapper(cls, self)

    @property
    def root(self):
        return self._root

    @property
    def serialized_model(self):
        return serializer.serialize_model(self._root, self.executor)[0]

    @property
    def preserve_exception(self):
        return self._preserve_exception

    @preserve_exception.setter
    def preserve_exception(self, value):
        self._preserve_exception = value

    def session(self):
        return helpers.with_object_store(self.executor.object_store)
